"""
High-level plotting routines
Also contains a dictionary for the valid plot types.
All plot types need to be imported and added to the plot_types dictionary in order to be used.
"""
import logging

from cis.plotting.comparativescatter import ComparativeScatter
from cis.plotting.contourplot import ContourPlot, ContourfPlot
from cis.plotting.heatmap import Heatmap
from cis.plotting.histogram import Histogram
from cis.plotting.histogram2d import Histogram2D
from cis.plotting.lineplot import LinePlot
from cis.plotting.scatterplot import ScatterPlot, ScatterPlot2D
from cis.plotting.taylor import Taylor
from cis.plugin import get_all_subclasses
import cartopy.crs
import six

colors = ['r', 'g', 'b', 'c', 'm', 'y', 'k']


plot_types = {"contour": ContourPlot,
              "contourf": ContourfPlot,
              "heatmap": Heatmap,
              "line": LinePlot,
              "scatter": ScatterPlot,
              "scatter2d": ScatterPlot2D,
              "comparativescatter": ComparativeScatter,
              "histogram": Histogram,
              "histogram2d": Histogram2D,
              "taylor": Taylor}


# Available projections from Cartopy
projections = {cls.__name__: cls for cls in get_all_subclasses(cartopy.crs.Projection, "cartopy.crs")}


def format_units(units):
    """
    Optionally put brackets around a units string
    :param units: The units of a variable, as a string
    :return: The units surrounding brackets, or the empty string if no units given
    """
    if "since" in str(units):
        # Assume we are on a time if the units contain since.
        return ""
    elif units:
        return "(" + str(units) + ")"
    else:
        return ""


def get_label(common_data, units=True):
    """
    Get a nicely formatted label from a CommonData object
    :param CommonData common_data:
    :param bool units: Include units in label? Default True
    :return string: A label
    """
    name = common_data.name() or ""
    if units:
        if name:
            name += " "
        name += format_units(common_data.units)
    return name


def _try_coord(data, coord_dict):
    """
    Try and find a single (extended) coord in a CommonData object using the dictionary provided
    :param CommonData data:
    :param dict coord_dict: Kwargs to look for coord
    :return: A single Coord or None if none can be found
    """
    import cis.exceptions as cis_ex
    import iris.exceptions as iris_ex
    try:
        coord = data.coord(**coord_dict)
    except (iris_ex.CoordinateNotFoundError, cis_ex.CoordinateNotFoundError):
        coord = None
    else:
        if not len(coord.points) > 1:
            # Don't guess a scalar coord
            coord = None
    return coord


def get_axis(d, axis, name=None):
    """
    Guess the best Coord to use for the axis
    :param CommonData d:
    :param string axis:
    :param string name:
    :return Coord: A single Coord
    """
    coord = _try_coord(d, dict(name_or_coord=name)) or _try_coord(d, dict(standard_name=name)) \
            or _try_coord(d, dict(axis=axis))

    # This is primarily for gridded data, but for Ungridded Data should just pick out the first Coord in the list.
    if coord is None:
        for c in d.coords():
            if axis == 'X' or len(d.shape) == 1:
                if c.shape[0] == d.shape[0]:
                    coord = c
                    break
            elif axis == 'Y' and len(d.shape) > 1:
                if c.shape[1] == d.shape[1]:
                    coord = c
                    break

    logging.info("Plotting " + coord.name() + " on the {} axis".format(axis))

    return coord


def set_x_axis_as_time(ax):
    """
    Nicely format a time axis
    """
    from matplotlib import ticker
    from matplotlib.dates import num2date

    def format_datetime(x, pos=None):
        # use iosformat rather than strftime as strftime can't handle dates before 1900 - the output is the same
        date_time = num2date(x)
        day_min, day_max = ax.get_xlim()
        day_range = day_max - day_min
        if day_range < 1 and date_time.second == 0:
            return "%02d" % date_time.hour + ':' + "%02d" % date_time.minute
        elif day_range < 1:
            return "%02d" % date_time.hour + ':' + "%02d" % date_time.minute + ':' + "%02d" % date_time.second
        elif day_range > 5:
            return str(date_time.date())
        else:
            return date_time.isoformat(' ')

    # Setup conversion of our standard time
    ax.xaxis.set_major_formatter(ticker.FuncFormatter(format_datetime))
    # Then just let matplotlib format it nicely...
    ax.get_figure().autofmt_xdate()


def _get_extent(ax):
    """
     Calculates the diagonal extent of plot area in Km
    :return: The diagonal size of the plot in Km
    """
    from cis.utils import haversine
    x0, x1, y0, y1 = ax.get_extent()
    return haversine(y0, x0, y1, x1)


def _test_natural_earth_available():
    """
    Test whether we can download the natural earth cartographies.
    :return: Can we access natural earth?
    """
    from cartopy.io.shapereader import natural_earth
    from six.moves.urllib.error import HTTPError
    try:
        natural_earth_available = natural_earth()
    except HTTPError:
        natural_earth_available = False
    return natural_earth_available


def drawcoastlines(ax, coastlinescolour):
    """
    Adds coastlines to a plot.
    There are three levels of resolution used based on the spatial scale of the plot. These are determined using
    values determined by eye for bluemarble and the coastlines independently.
    """
    coastline_scales = [(0, '110m'), (500, '50m'), (100, '10m')]

    ext = _get_extent(ax)

    if _test_natural_earth_available():
        coastline_res = coastline_scales[0][1]
        for scale, res in coastline_scales[1:]:
            if scale > ext:
                coastline_res = res

        ax.coastlines(color=coastlinescolour, resolution=coastline_res)
    else:
        logging.warning('Unable to access the natural earth topographies required for plotting coastlines. '
                        'Check internet connectivity and try again')


def drawbluemarble(ax):
    """
    Adds nasa blue marble back ground to a plot
    There are three levels of resolution used based on the spatial scale of the plot. These are determined using
    values determined by eye for bluemarble and the coastlines independently.
    """
    from matplotlib.image import imread
    import cartopy.crs as ccrs
    import os.path as path
    from cis.utils import no_autoscale

    source_proj = ccrs.PlateCarree()

    bluemarble_scales = [(0, 'raster/world.topo.bathy.200407.3x1350x675.png'),
                         (5000, 'raster/world.topo.bathy.200407.3x2700x1350.png'),
                         (2500, 'raster/world.topo.bathy.200407.3x5400x2700.png')]

    ext = _get_extent(ax)

    # Search for the right resolution
    bluemarble_res = bluemarble_scales[0][1]
    for scale, res in bluemarble_scales[1:]:
        if scale > ext:
            bluemarble_res = res

    img = imread(path.join(path.dirname(path.realpath(__file__)), bluemarble_res))
    # Don't change the scale of the plot
    with no_autoscale(ax):
        ax.imshow(img, origin='upper', transform=source_proj, extent=[-180, 180, -90, 90])


def get_best_map_ticks(ax, transform=None):
    """
    Use the matplotlib.ticker class to automatically set nice values for the major and minor ticks.
    Log axes generally come out nicely spaced without needing manual intervention. For particularly narrow latitude
    vs longitude plots the ticks can come out overlapped, so an exception is included to deal with this.
    """
    from matplotlib.ticker import MaxNLocator

    max_x_bins = 9
    max_y_bins = 7  # as plots are wider rather than taller

    lon_steps = [1, 3, 6, 9, 10]
    lat_steps = [1, 3, 6, 9, 10]
    variable_step = [1, 2, 4, 5, 10]

    xmin, xmax, ymin, ymax = ax.get_extent(crs=transform)
    # ymin, ymax = ax.get_ylim()

    if (xmax - xmin) < 5:
        lon_steps = variable_step
    if (ymax - ymin) < 5:
        lat_steps = variable_step

    # We need to make a special exception for particularly narrow and wide plots, which will be lat vs lon
    # preserving the aspect ratio. This gives more options for the spacing to try and find something that can use
    # the maximum number of bins.
    if (ymax - ymin) > 2.2 * (xmax - xmin):
        max_x_bins = 4
        max_y_bins = 11
    elif (xmax - xmin) > 2.2 * (ymax - ymin):
        max_x_bins = 14
        max_y_bins = 4

    lon_locator = MaxNLocator(nbins=max_x_bins, steps=lon_steps)
    lat_locator = MaxNLocator(nbins=max_y_bins, steps=lat_steps)
    lon_ticks = lon_locator.tick_values(xmin, xmax)
    # Prune any large longitude ticks
    lon_ticks = [t for t in lon_ticks if abs(t) <= 360.0]
    lat_ticks = lat_locator.tick_values(ymin, ymax)
    # Prune any large latitude ticks
    lat_ticks = [t for t in lat_ticks if abs(t) <= 90.0]
    return lon_ticks, lat_ticks


def set_map_ticks(ax, xticks, yticks, transform=None):
    from cartopy.mpl.ticker import LongitudeFormatter, LatitudeFormatter
    import cartopy.crs as ccrs

    if isinstance(ax.projection, ccrs._RectangularProjection):
        ax.set_xticks(xticks, crs=transform)
        ax.xaxis.set_major_formatter(LongitudeFormatter())

        ax.set_yticks(yticks, crs=transform)
        ax.yaxis.set_major_formatter(LatitudeFormatter())
    else:
        logging.debug("Not setting map ticks for non-rectangular projection. Setting gridlines instead.")
        ax.gridlines()


def add_color_bar(ax, mappable, vstep, logv, cbarscale, cbarorient, cbarlabel):
    """
    Adds a colour bar to a plot
    Allows specifying of tick spacing and orientation
    """
    import matplotlib.pyplot as plt
    from matplotlib.ticker import MultipleLocator, LogFormatterMathtext

    cbar_kwargs = {}

    if vstep is not None:
        cbar_kwargs['ticks'] = MultipleLocator(vstep)

    if logv:
        cbar_kwargs['format'] = LogFormatterMathtext(10, labelOnlyBase=False)
    #
    if cbarscale is not None:
        cbar_kwargs['shrink'] = cbarscale

    if cbarorient is not None:
        cbar_kwargs['orientation'] = cbarorient

    cbar = plt.colorbar(mappable, ax=ax, **cbar_kwargs)

    if not logv:
        cbar.formatter.set_scientific(True)
        cbar.formatter.set_powerlimits((-3, 3))
        cbar.update_ticks()

    cbar.set_label(cbarlabel)


def basic_plot(data, how=None, ax=None, xaxis=None, yaxis=None, projection=None, central_longitude=0.0,
               label=None, *args, **kwargs):
    """
    Plot a single data object using the method specified

    :param CommonData data: The data to plot
    :param string how: The method to use, one of:  "contour", "contourf", "heatmap", "line", "scatter", "scatter2d",
    "comparativescatter", "histogram", "histogram2d" or "taylor"
    :param Axes ax: A matplotlib axes on which to draw the plot
    :param Coord or CommonData xaxis: The data to plot on the x axis
    :param Coord or CommonData yaxis: The data to plot on the y axis
    :param string or cartopy.crs.Projection projection: The projection to use for map plots (default is PlateCaree)
    :param float central_longitude: The central longitude to use for PlateCaree (if no other projection specified)
    :param string label: A label for the data. This is used for the title, colorbar or legend depending on plot type
    :param args: Other plot-specific args
    :param kwargs: Other plot-specific kwargs
    :return APlot and Axes: The APlot instance and the matplotlib Axes on which the plot was drawn
    """
    import cartopy.crs as ccrs
    from cartopy.mpl.geoaxes import GeoAxes
    import matplotlib.pyplot as plt
    from cis.data_io.common_data import CommonData
    from cis.data_io.gridded_data import GriddedData
    from cis.utils import squeeze
    from cis.plotting.genericplot import GenericPlot

    # Remove any extra gridded data dimensions
    if isinstance(data, GriddedData):
        data = squeeze(data)

    if isinstance(xaxis, CommonData):
        if how is not None:
            if how not in ['comparativescatter', 'histogram2d']:
                raise ValueError("Invalid plot type with xaxis as a CommonData object. Either use comparativescatter, "
                                 "histogram2d or use a Coord (or None) for the xaxis argument.")
        else:
            how = 'comparativescatter'
        xaxis_coord = xaxis
    else:
        xaxis_coord = get_axis(data, 'X', xaxis)

    yaxis_coord = get_axis(data, 'Y', yaxis)

    how = how or data._get_default_plot_type(xaxis_coord.standard_name == 'longitude'
                                             and yaxis_coord.standard_name == 'latitude')

    try:
        plt_type = plot_types[how]
    except KeyError:
        raise ValueError("Invalid plot type, must be one of: {}".format(plot_types.keys()))

    if issubclass(plt_type, GenericPlot) and yaxis is not None:
        # If we have a line or scatter plot and the user has specified a y-axis, then this should be the data
        data = yaxis_coord
    else:
        # Otherwise, it's just another kwarg
        kwargs['yaxis'] = yaxis_coord

    # Set a nice default label
    label = get_label(data) if label is None else label

    plot = plot_types[how](data, xaxis=xaxis_coord, label=label, *args, **kwargs)

    subplot_kwargs = {}
    if plot.is_map():
        if projection is None:
            projection = ccrs.PlateCarree(central_longitude=central_longitude)
        elif isinstance(projection, six.string_types):
            projection = projections[projection]()
        plot.mplkwargs['transform'] = ccrs.PlateCarree()
        subplot_kwargs['projection'] = projection
        # Monkey-patch the nasabluemarble method onto the axis
        GeoAxes.bluemarble = drawbluemarble

    if ax is None:
        _, ax = plt.subplots(subplot_kw=subplot_kwargs)

    # Make the plot
    plot(ax)

    # Any post-processing
    if xaxis_coord.standard_name == 'time' and how not in ['comparativescatter', 'histogram2d', 'histogram']:
        set_x_axis_as_time(ax)

    if yaxis_coord.standard_name == 'air_pressure':
        ax.invert_yaxis()

    return plot, ax


def multilayer_plot(data_list, how=None, ax=None, yaxis=None, layer_opts=None, *args, **kwargs):
    """
    Plot multiple data objects using the method specified

    :param list data_list: A list of CommonData objects to plot
    :param string how: The method to use, one of:  "contour", "contourf", "heatmap", "line", "scatter", "scatter2d",
    "comparativescatter", "histogram", "histogram2d" or "taylor"
    :param Axes ax: A matplotlib axes on which to draw the plot
    :param Coord or CommonData xaxis: The data to plot on the x axis
    :param Coord or CommonData yaxis: The data to plot on the y axis
    :param list layer_opts: A list of keyword dictionaries to pass to each layer of the plot.
    :param args: Other plot-specific args to pass to all plots
    :param kwargs: Other plot-specific kwargs to pass to all plots
    :return APlot and Axes: The APlot instance and the matplotlib Axes on which the plot was drawn
    """
    layer_opts = [{} for i in data_list] if layer_opts is None else layer_opts
    if len(layer_opts) != len(data_list):
        raise ValueError("One layer-options keyword dictionary must be supplied for each data item, or none at all.")

    if how in ['comparativescatter', 'histogram2d']:
        if len(data_list) != 2:
            raise ValueError("Exactly two data objects must be provided for comparative plots")
        layer_kwargs = dict(list(kwargs.items()) + list(layer_opts[0].items()))
        plot, ax = basic_plot(data_list[1], how, ax, xaxis=data_list[0], *args, **layer_kwargs)
    elif how == 'taylor':
        plot, ax = taylor_plot(data_list, ax, layer_opts, *args, **kwargs)
    else:
        if not isinstance(yaxis, list):
            yaxis = [yaxis for i in data_list]

        for d, y, opts in zip(data_list, yaxis, layer_opts):
            layer_kwargs = dict(list(kwargs.items()) + list(opts.items()))
            how = layer_kwargs.pop('type', how)
            label = layer_kwargs.pop('label', get_label(d, units=False))
            plot, ax = basic_plot(d, how, ax, yaxis=y, label=label, *args, **layer_kwargs)

        legend = ax.legend(loc="best")
        if legend is not None:
            legend.draggable(state=True)

    return plot, ax


def taylor_plot(data_list, ax=None, layer_opts=None, *args, **kwargs):
    """
    Construct a Taylor diagram from the data list provided. Taylor plots are a bit special!
     Layer_opts are parsed for itemstyle and color which are then combined to be passed to the plotting routines. This
     allows reuse of familiar command line kwargs while maintaining a simple API.

    :param list data_list: List of CommonData objexts
    :param ax: Optional axis - although we don't use a standard axis for this plot so using the default is strongly
    recommended
    :param list layer_opts: A list of dictionaries optionally containing labels, itemstyles and colors for each data object
    :param list args: Optional extra arguments
    :param dict kwargs: Optional extra keyword arguments
    :return: Taylor plot and axes instances
    """
    import matplotlib.pyplot as plt
    from cis.plotting.taylor import ArcCosTransform
    from matplotlib.projections import PolarAxes
    from matplotlib.transforms import IdentityTransform, blended_transform_factory
    import mpl_toolkits.axisartist.floating_axes as floating_axes

    _ = kwargs.pop('central_longitude', None)  # In case the Plotter has added it...

    layer_opts = [{} for i in data_list] if layer_opts is None else layer_opts
    labels = [layer_opt.pop('label', None) for layer_opt in layer_opts]

    # Pull together markers from the layer_opts
    markers = [layer_opt.pop('itemstyle', None) for layer_opt in layer_opts]
    if all(m is None for m in markers):
        # If all markers are None then just set the list to None
        markers = None
    elif any(m is None for m in markers):
        # If not all are None, but some are then we have a problem
        raise ValueError("If any markers are set then a marker must be set for every dataset")

    # Pull together colors from the layer_opts
    colors = [layer_opt.pop('color', None) for layer_opt in layer_opts]
    if all(c is None for c in colors):
        # If all markers are None then just set the list to None
        colors = None
    elif any(c is None for c in colors):
        # If not all are None, but some are then we have a problem
        raise ValueError("If any markers are set then a marker must be set for every dataset")
    kwargs['itemwidth'] = layer_opts[0].pop('itemwidth', None)
    plot = Taylor(data_list, labels, colors, markers, *args, **kwargs)

    if ax is None:
        fig = plt.figure()

        tr = blended_transform_factory(ArcCosTransform(), IdentityTransform()) + PolarAxes.PolarTransform()

        gh = floating_axes.GridHelperCurveLinear(tr, extremes=(plot.extend, 1., 0., plot.gammamax),
                                                 grid_locator1=None,
                                                 grid_locator2=None,
                                                 tick_formatter1=None,
                                                 tick_formatter2=None)
        ax = floating_axes.FloatingSubplot(fig, 1, 1, 1, grid_helper=gh)
        fig.add_subplot(ax)

    ax = plot(ax)

    return plot, ax



"""
Module containing hdf file utility functions for the SD object
"""
import logging
from cis.utils import listify
# Optional HDF import, if the module isn't found we defer raising ImportError until it is actually needed.
try:
    from pyhdf import SD
except ImportError:
    SD = None


def get_hdf_SD_file_variables(filename):
    """
    Get all the variables from an HDF SD file

    :param str filename: The filename of the file to get the variables from
    :return: An OrderedDict containing the variables from the file
    """
    if not SD:
        raise ImportError("HDF support was not installed, please reinstall with pyhdf to read HDF files.")

    variables = None

    try:
        # Open the file.
        datafile = SD.SD(filename)
        # List of required variable names.
        variables = datafile.datasets()
        # Close the file
        datafile.end()
    except:
        logging.error("Error while reading SD data")

    return variables


class HDF_SDS(object):
    """
    This class is used in place of the pyhdf.SD.SDS class to allow the file contents to be loaded at a later time
    rather than in this module read method (so that we can close the SD instances and free up file handles)
    """

    _sd = None
    _sds = None
    _filename = None
    _variable = None

    def __init__(self, filename, variable, start=None, count=None, stride=None):
        self._filename = filename
        self._variable = variable
        self._start = start
        self._count = count
        self._stride = stride

    def _open_sds(self):
        """
        Open the SDS file for reading
        """
        from pyhdf.SD import SD as SDS

        self._sd = SDS(self._filename)
        self._sds = self._sd.select(self._variable)

    def _close_sds(self):
        """
        Close the SDS file for reading

        NB: Exceptions thrown from here may hide an exception thrown in get(), info(), etc.
        """
        try:
            if self._sds is not None:
                self._sds.endaccess()
        finally:
            if self._sd is not None:
                self._sd.end()

    def get(self, start=None, count=None, stride=None):
        """
        Call pyhdf.SD.SDS.get(), opening and closing the file
        """
        if start is None:
            start = self._start
        if count is None:
            count = self._count
        if stride is None:
            stride = self._stride
        try:
            self._open_sds()
            data = self._sds.get(start, count, stride)
            return data
        finally:
            self._close_sds()

    def attributes(self):
        """
        Call pyhdf.SD.SDS.attributes(), opening and closing the file
        """
        try:
            self._open_sds()
            attributes = self._sds.attributes()
            return attributes
        finally:
            self._close_sds()

    def info(self):
        """
        Call pyhdf.SD.SDS.info(), opening and closing the file
        """
        try:
            self._open_sds()
            info = self._sds.info()
            return info
        finally:
            self._close_sds()

    def dimensions(self):
        """
        Call pyhdf.SD.SDS.dimensions(), opening and closing the file
        """
        from collections import OrderedDict
        try:
            self._open_sds()
            var_description = self._sd.datasets()[self._variable]
            return OrderedDict(zip(var_description[0], var_description[1]))
        finally:
            self._close_sds()


def read(filename, variables=None, datadict=None):
    """
    Reads SD from a HDF4 file into a dictionary.

    :param str filename: The name (with path) of the HDF file to read.
    :param iterable names: A sequence of variable (dataset) names to read from the
     file (default None, causing all variables to be read). The names must appear exactly as in in the HDF file.
    :param dict datadict: Optional dictionary to add data to, otherwise a new, empty dictionary is created
    :return: A dictionary containing data for requested variables. Missing data is replaced by NaN.
    """
    # Optional HDF import
    if not SD:
        raise ImportError("HDF support was not installed, please reinstall with pyhdf to read HDF files.")

    # List of required variable names.
    # Open the file.
    datafile = None
    try:
        datafile = SD.SD(filename)
        sd_variables = list(datafile.datasets().keys())
    finally:
        if datafile is not None:
            datafile.end()

    if variables is None:
        requested_sd_variables = sd_variables
    else:
        requested_sd_variables = set(listify(variables)).intersection(set(sd_variables))

    # Create dictionary to hold data arrays for returning.
    if datadict is None:
        datadict = {}

    # Get data.
    for variable in requested_sd_variables:
        datadict[variable] = HDF_SDS(filename, variable)

    return datadict


def get_data(sds):
    """
    Reads raw data from an SD instance.

    :param sds: The specific sds instance to read
    :return: A numpy array containing the raw data with missing data is replaced by NaN.
    """
    from cis.utils import create_masked_array_for_missing_data
    from cis.data_io.netcdf import apply_offset_and_scaling
    import numpy as np

    data = sds.get()
    attributes = sds.attributes()

    # Apply Fill Value
    missing_value = attributes.get('_FillValue', None)
    if missing_value is not None:
        data = create_masked_array_for_missing_data(data, missing_value)

    # Check for valid_range
    valid_range = attributes.get('valid_range', None)
    if valid_range is not None:
        data = np.ma.masked_outside(data, *valid_range)

    # Offsets and scaling.
    add_offset = attributes.get('add_offset', 0.0)
    scale_factor = attributes.get('scale_factor', 1.0)
    logging.warning("Applying standard offset and scaling for dataset - this may not be appropriate for HDF_EOS data!")
    data = apply_offset_and_scaling(data, add_offset=add_offset, scale_factor=scale_factor)

    return data


def get_metadata(sds):
    from cis.data_io.ungridded_data import Metadata

    name = sds.info()[0]

    misc = sds.attributes()
    long_name = misc.pop('long_name', '')
    units = misc.pop('units', '')
    factor = misc.pop('scale_factor', None)
    offset = misc.pop('add_offset', None)
    missing = misc.pop('_FillValue', None)

    shape = sds.info()[2]

    metadata = Metadata(name=name, long_name=long_name, shape=shape, units=units,
                        factor=factor, offset=offset, missing_value=missing, misc=misc)

    return metadata



"""
A contour plot, filled or not.
"""
from .genericplot import Generic2DPlot


class ContourPlot(Generic2DPlot):

    def __init__(self, packed_data, contnlevels=7, vstep=None,
                 contlevels=None, contlabel=False, contwidth=None, cont_label_kwargs=None, *args, **kwargs):
        """

        :param packed_data:
        :param int contnlevels: The number of contour levels to plot (default is 7)
        :param float vstep: The step between contour levels
        :param list contlevels: A list of contour levels to plot
        :param bool contlabel: Plot contour labels (default is False)
        :param int contwidth: The thickness of the contour lines
        :param dict cont_label_kwargs: A dictionary of contour label keywork args (e.g. format)
        :param kwargs: Other keyword args to pass to plot
        """
        super(ContourPlot, self).__init__(packed_data, *args, **kwargs)
        self.filled = False
        if contlevels:
            self.levels = contlevels
        elif self.vstep:
            self.levels = (kwargs.get('vmax', self.data.max()) -
                           kwargs.get('vmin', self.data.min())) / vstep
        else:
            self.levels = contnlevels

        self.contlabel = contlabel
        self.contwidth = contwidth
        self.cont_label_kwargs = cont_label_kwargs if cont_label_kwargs is not None else {}

    def __call__(self, ax):
        from matplotlib import ticker

        # Set the options specific to a datagroup with the contour type
        mplkwargs = self.mplkwargs

        mplkwargs["colors"] = self.color

        mplkwargs["linewidths"] = self.contwidth

        if self.logv:
            mplkwargs['locator'] = ticker.LogLocator()

        contour_type = ax.contourf if self.filled else ax.contour

        self.mappable = contour_type(self.x, self.y, self.data, self.levels, **mplkwargs)

        if self.contlabel:
            ax.clabel(self.mappable, **self.cont_label_kwargs)

        super(ContourPlot, self).__call__(ax)


class ContourfPlot(ContourPlot):

    def __init__(self, packed_data, *args, **kwargs):
        super(ContourfPlot, self).__init__(packed_data, *args, **kwargs)
        self.filled = True



import datetime as dt
import pandas as pd

from hamcrest import *
from nose.tools import istest, eq_
import numpy as np
from cis.collocation.kdtree import KDTree
from cis.time_util import cis_standard_time_unit
import cis.data_io.gridded_data as gridded_data
from cis.data_io.hyperpoint import HyperPoint, HyperPointList
from cis.data_io.ungridded_data import UngriddedData
from cis.test.util import mock
from cis.collocation.col_implementations import (GeneralUngriddedCollocator, nn_horizontal_only, DummyConstraint,
                                                 SepConstraintKdtree, make_coord_map)
from cis.collocation.haversinedistancekdtreeindex import HaversineDistanceKDTreeIndex


def is_collocated(data1, data2):
    """
    Checks whether two datasets share all of the same points, this might be useful
    to determine if collocation is necessary or completed successfully
    """
    collocated = len(data1) == len(data2)
    if collocated:
        for i, point1 in enumerate(data1):
            collocated = point1.same_point_in_space_and_time(data2[i])
            if not collocated:
                return collocated
    return collocated


class Test_nn_horizontal_kdtree(object):
    @istest
    def test_basic_col_in_2d(self):
        # lat: -10 to 10 step 5; lon -5 to 5 step 5
        ug_data = mock.make_regular_2d_ungridded_data()
        sample_points = UngriddedData.from_points_array(
            [HyperPoint(lat=1.0, lon=1.0), HyperPoint(lat=4.0, lon=4.0), HyperPoint(lat=-4.0, lon=-4.0)])
        col = GeneralUngriddedCollocator(fill_value=-999)
        new_data = col.collocate(sample_points, ug_data, SepConstraintKdtree(), nn_horizontal_only())[0]
        eq_(new_data.data[0], 8.0)
        eq_(new_data.data[1], 12.0)
        eq_(new_data.data[2], 4.0)

    @istest
    def test_already_collocated_in_col_ungridded_to_ungridded_in_2d(self):
        ug_data = mock.make_regular_2d_ungridded_data()
        # This point already exists on the cube with value 5 - which shouldn't be a problem
        sample_points = UngriddedData.from_points_array([HyperPoint(0.0, 0.0)])
        col = GeneralUngriddedCollocator(fill_value=-999)
        new_data = col.collocate(sample_points, ug_data, SepConstraintKdtree(), nn_horizontal_only())[0]
        eq_(new_data.data[0], 8.0)

    @istest
    def test_coordinates_exactly_between_points_in_col_ungridded_to_ungridded_in_2d(self):
        """
        This works out the edge case where the points are exactly in the middle or two or more datapoints.
        The nn_horizontal algorithm will start with the first point as the nearest and iterates through the
        points finding any points which are closer than the current closest. If two distances were exactly the same
        you would expect the first point to be chosen. This doesn't seem to always be the case but is probably
        down to floating points errors in the haversine calculation as these test points are pretty close
        together. This test is only really for documenting the behaviour for equidistant points.
        """
        ug_data = mock.make_regular_2d_ungridded_data()
        sample_points = UngriddedData.from_points_array(
            [HyperPoint(2.5, 2.5), HyperPoint(-2.5, 2.5), HyperPoint(2.5, -2.5),
             HyperPoint(-2.5, -2.5)])
        col = GeneralUngriddedCollocator(fill_value=-999)
        new_data = col.collocate(sample_points, ug_data, SepConstraintKdtree(), nn_horizontal_only())[0]
        eq_(new_data.data[0], 11.0)
        eq_(new_data.data[1], 5.0)
        eq_(new_data.data[2], 10.0)
        eq_(new_data.data[3], 4.0)

    @istest
    def test_coordinates_outside_grid_in_col_ungridded_to_ungridded_in_2d(self):
        ug_data = mock.make_regular_2d_ungridded_data()
        sample_points = UngriddedData.from_points_array(
            [HyperPoint(5.5, 5.5), HyperPoint(-5.5, 5.5), HyperPoint(5.5, -5.5),
             HyperPoint(-5.5, -5.5)])
        col = GeneralUngriddedCollocator(fill_value=-999)
        new_data = col.collocate(sample_points, ug_data, SepConstraintKdtree(), nn_horizontal_only())[0]
        eq_(new_data.data[0], 12.0)
        eq_(new_data.data[1], 6.0)
        eq_(new_data.data[2], 10.0)
        eq_(new_data.data[3], 4.0)


class TestSepConstraint(object):
    @istest
    def test_horizontal_constraint_in_2d(self):
        from cis.data_io.Coord import Coord, CoordList
        from cis.data_io.ungridded_data import Metadata
        ug_data = mock.make_regular_2d_ungridded_data()
        ug_data_points = ug_data.as_data_frame(time_index=False, name='vals').dropna(axis=1)
        sample_points = UngriddedData(np.array([0.0]), Metadata(),
                                      CoordList([Coord(np.array([7.5]), Metadata(standard_name='latitude')),
                                                 Coord(np.array([-2.5]), Metadata(standard_name='longitude'))]))
        sample_points_view = sample_points.as_data_frame(time_index=False, name='vals').dropna(axis=1)
        # sample_point = HyperPoint(lat=7.5, lon=-2.5)
        # sample_points = HyperPointList([sample_point])
        coord_map = None

        # One degree near 0, 0 is about 110km in latitude and longitude, so 300km should keep us to within 3 degrees
        #  in each direction
        constraint = SepConstraintKdtree(h_sep=400)

        index = HaversineDistanceKDTreeIndex()
        index.index_data(sample_points, ug_data_points, coord_map, leafsize=2)
        constraint.haversine_distance_kd_tree_index = index

        # This should leave us with 4 points
        ref_vals = np.array([10, 11, 13, 14])

        indices = constraint.haversine_distance_kd_tree_index.find_points_within_distance_sample(sample_points_view, 400)

        new_vals = ug_data.data.flat[indices]

        eq_(ref_vals.size, len(new_vals[0]))
        assert (np.equal(ref_vals, new_vals).all())

    @istest
    def test_horizontal_constraint_in_2d_with_missing_values(self):
        ug_data = mock.make_regular_2d_ungridded_data_with_missing_values()
        ug_data_points = ug_data.as_data_frame(time_index=False, name='vals').dropna(axis=0)
        coord_map = None

        # One degree near 0, 0 is about 110km in latitude and longitude, so 300km should keep us to within 3 degrees
        #  in each direction
        constraint = SepConstraintKdtree(h_sep=400)

        index = HaversineDistanceKDTreeIndex()
        sample_points = None  # Not used
        index.index_data(sample_points, ug_data_points, coord_map, leafsize=2)
        constraint.haversine_distance_kd_tree_index = index

        for i, sample_point in ug_data.as_data_frame(time_index=False, name='vals').iterrows():
            new_points = constraint.constrain_points(sample_point, ug_data_points)
            new_vals = new_points.vals
            if np.isnan(sample_point.vals):
                ref_vals = np.array([])
            else:
                ref_vals = np.array([sample_point.vals])

            eq_(ref_vals.size, new_vals.size)
            assert (np.equal(ref_vals, new_vals).all())

    @istest
    def test_horizontal_constraint_in_4d(self):
        ug_data = mock.make_regular_4d_ungridded_data()
        ug_data_points = ug_data.as_data_frame(time_index=False, name='vals').dropna(axis=1)
        sample_points = pd.DataFrame(data={'longitude': [0.0], 'latitude': [0.0], 'altitude': [50.0],
                                              'time': [dt.datetime(1984, 8, 29)]})
        coord_map = None

        # Constraint distance selects the central three points.
        constraint = SepConstraintKdtree(h_sep=1000)

        index = HaversineDistanceKDTreeIndex()
        index.index_data(sample_points, ug_data_points, coord_map)
        constraint.haversine_distance_kd_tree_index = index

        # This should leave us with 30 points
        ref_vals = np.reshape(np.arange(50) + 1.0, (10, 5))[:, 1:4].flatten()

        new_points = constraint.constrain_points(sample_points.iloc[0], ug_data_points)
        new_vals = np.sort(new_points.vals)

        eq_(ref_vals.size, new_vals.size)
        assert (np.equal(ref_vals, new_vals).all())

    @istest
    def test_all_constraints_in_4d(self):
        ug_data = mock.make_regular_4d_ungridded_data()
        ug_data_points = ug_data.as_data_frame(time_index=False, name='vals').dropna(axis=1)
        sample_point = pd.DataFrame(data={'longitude': [0.0], 'latitude': [0.0], 'altitude': [50.0],
                                          'air_pressure': [50.0],
                                          'time': [cis_standard_time_unit.date2num(dt.datetime(1984, 8, 29))]}).iloc[0]
        # One degree near 0, 0 is about 110km in latitude and longitude, so 300km should keep us to within 3 degrees
        #  in each direction
        h_sep = 1000
        # 15m altitude separation
        a_sep = 15
        # 1 day (and a little bit) time separation
        t_sep = 'P1dT1M'
        # Pressure constraint is 50/40 < p_sep < 60/50
        p_sep = 1.22

        constraint = SepConstraintKdtree(h_sep=h_sep, a_sep=a_sep, p_sep=p_sep, t_sep=t_sep)

        index = HaversineDistanceKDTreeIndex()
        index.index_data(None, ug_data_points, None)
        constraint.haversine_distance_kd_tree_index = index

        # This should leave us with 9 points: [[ 22, 23, 24]
        #                                      [ 27, 28, 29]
        #                                      [ 32, 33, 34]]
        ref_vals = np.array([27., 28., 29., 32., 33., 34.])

        new_points = constraint.constrain_points(sample_point, ug_data_points)
        new_vals = np.sort(new_points.vals)

        eq_(ref_vals.size, new_vals.size)
        assert (np.equal(ref_vals, new_vals).all())

    def get_max_depth(self, node, depth):
        if isinstance(node, KDTree.leafnode):
            return depth
        return max(self.get_max_depth(node.less, depth + 1), self.get_max_depth(node.greater, depth + 1))

    @istest
    def test_horizontal_constraint_in_2d_when_lats_are_the_same_produces_a_balanced_tree(self):
        ug_data = mock.make_regular_2d_ungridded_data(lat_dim_length=1001, lat_max=10, lat_min=10)
        ug_data_points = ug_data.as_data_frame(time_index=False, name='vals').dropna(axis=1)
        sample_points = pd.DataFrame(data={'longitude': [-2.5], 'latitude': [7.5]})
        coord_map = None

        # One degree near 0, 0 is about 110km in latitude and longitude, so 300km should keep us to within 3 degrees
        #  in each direction
        constraint = SepConstraintKdtree(h_sep=400)

        index = HaversineDistanceKDTreeIndex()
        index.index_data(sample_points, ug_data_points, coord_map, leafsize=2)

        depth = self.get_max_depth(index.index.tree, 0)

        assert_that(depth, is_(2), "Depth is 2, there are three unique values -10, 0, 10")


class TestSepConstraintWithoutHorizontalSeparation(object):
    """Tests that SepConstraintKdtree behaves as an unoptimized constraint for non-spatial separations
    if the spatial separation parameter is not specified.
    """

    @istest
    def test_alt_constraint_in_4d(self):
        from cis.collocation.col_implementations import SepConstraintKdtree
        import datetime as dt
        import numpy as np

        ug_data = mock.make_regular_4d_ungridded_data()
        ug_data_points = ug_data.as_data_frame(time_index=False, name='vals').dropna(axis=1)
        sample_point = pd.Series({'longitude': [0.0], 'latitude': [0.0], 'altitude':[50.0],
                                  'time': [cis_standard_time_unit.date2num(dt.datetime(1984, 8, 29))]})
        # 15m altitude separation
        a_sep = 15

        constraint = SepConstraintKdtree(a_sep=a_sep)

        # This should leave us with 15 points:  [ 21.  22.  23.  24.  25.]
        #                                       [ 26.  27.  28.  29.  30.]
        #                                       [ 31.  32.  33.  34.  35.]
        ref_vals = np.array([21., 22., 23., 24., 25., 26., 27., 28., 29., 30., 31., 32., 33., 34., 35.])

        new_points = constraint.constrain_points(sample_point, ug_data_points)
        new_vals = new_points.vals

        eq_(ref_vals.size, new_vals.size)
        assert (np.equal(ref_vals, new_vals).all())

    @istest
    def test_time_constraint_in_4d(self):
        from cis.collocation.col_implementations import SepConstraintKdtree
        import datetime as dt
        import numpy as np

        ug_data = mock.make_regular_4d_ungridded_data()
        ug_data_points = ug_data.as_data_frame(time_index=False, name='vals').dropna(axis=1)
        sample_point = pd.Series({'longitude': [0.0], 'latitude': [0.0], 'altitude':[50.0],
                                  'time': [cis_standard_time_unit.date2num(dt.datetime(1984, 8, 29))]})

        # 1 day (and a little bit) time seperation
        constraint = SepConstraintKdtree(t_sep='P1dT1M')

        # This should leave us with 30 points
        ref_vals = np.reshape(np.arange(50) + 1.0, (10, 5))[:, 1:4].flatten()

        new_points = constraint.constrain_points(sample_point, ug_data_points)
        new_vals = new_points.vals

        eq_(ref_vals.size, new_vals.size)
        assert (np.equal(ref_vals, new_vals).all())

    @istest
    def test_pressure_constraint_in_4d(self):
        from cis.collocation.col_implementations import SepConstraintKdtree
        import datetime as dt
        import numpy as np

        ug_data = mock.make_regular_4d_ungridded_data()
        ug_data_points = ug_data.as_data_frame(time_index=False, name='vals').dropna(axis=1)
        sample_point = pd.Series({'longitude': [0.0], 'latitude': [0.0], 'altitude':[50.0], 'air_pressure': [24.0],
                                  'time': [cis_standard_time_unit.date2num(dt.datetime(1984, 8, 29))]})
        constraint = SepConstraintKdtree(p_sep=2)

        # This should leave us with 20 points:  [  6.   7.   8.   9.  10.]
        #                                       [ 11.  12.  13.  14.  15.]
        #                                       [ 16.  17.  18.  19.  20.]
        #                                       [ 21.  22.  23.  24.  25.]
        ref_vals = np.array([6., 7., 8., 9., 10., 11., 12., 13., 14., 15., 16., 17., 18., 19., 20., 21., 22., 23.,
                             24., 25.])

        new_points = constraint.constrain_points(sample_point, ug_data_points)
        new_vals = new_points.vals

        eq_(ref_vals.size, new_vals.size)
        assert (np.equal(ref_vals, new_vals).all())


if __name__ == '__main__':
    import nose

    nose.runmodule()



import re
from cis.time_util import cis_standard_time_unit
import argparse


def parse_datetimestr_to_std_time(s):
    import dateutil.parser as du
    return cis_standard_time_unit.date2num(du.parse(s))


def _parse_datetime(s):
    """Parse a date/time string.

    The string should be in an ISO 8601 format except that the date and time
    parts may be separated by a space or colon instead of T.
    """
    import dateutil.parser as du
    return du.parse(s)


def _parse_partial_datetime(dt_string):
    """Parse a partial date/time string.

    The string should be in an ISO 8601 format except that the date and time
    parts may be separated by a space or colon instead of T.
    :param dt_string: String to parse
    :return: list of datetime components
    :raise ValueError: if the string cannot be parsed as a date/time
    """
    from cis.time_util import PartialDateTime

    # Separate date and time at first character that is one of 'T', ' ' or ':'.
    match = re.match(r'(?P<date>[^T :]+)(?:[T :])(?P<time>.+)$', dt_string)
    if match is not None:
        date_str = match.group('date')
        time_components = [int(x) for x in match.group('time').split(':')]
    else:
        date_str = dt_string
        time_components = []

    date_components = [int(x) for x in date_str.split('-')]

    dt_components = list(date_components)
    if ((len(date_components) > 3) or (len(time_components) > 3) or
            (len(date_components) < 3) and (len(time_components) > 0)):
        raise ValueError()
    else:
        dt_components.extend(time_components)

    return PartialDateTime(*dt_components)


def parse_partial_datetime(dt_string, name, parser):
    """Parse a partial date/time string from the command line, reporting parse errors.

    The string should be in an ISO 8601 format except that the date and time
    parts may be separated by a space or colon instead of T.
    :param dt_string: String to parse
    :param name:      A description of the argument used for error messages
    :param parser:    The parser used to report errors
    :return: datetime value
    """
    if dt_string == 'None' or dt_string is None:
        return None
    try:
        dt = _parse_partial_datetime(dt_string)
    except ValueError:
        parser.error("'" + dt_string + "' is not a valid " + name)
        dt = None
    return dt


def parse_datetime(dt_string, name, parser):
    """Parse a date/time string from the command line, reporting parse errors.

    The string should be in an ISO 8601 format except that the date and time
    parts may be separated by a space or colon instead of T.
    :param dt_string: String to parse
    :param name:      A description of the argument used for error messages
    :param parser:    The parser used to report errors
    :return: datetime value
    """
    if dt_string == 'None' or dt_string is None:
        return None
    try:
        dt = _parse_datetime(dt_string)
    except ValueError:
        parser.error("'" + dt_string + "' is not a valid " + name)
        dt = None
    return dt


def _parse_datetime_delta(dt_string):
    """Parse a date/time delta string into years, months, days, hours, minutes and seconds.

    :param dt_string: String to parse, for example 'PY2M3DT4H5M6S' (ISO 8601)
    :return: Named tuple 'date_delta' containing, 'year', 'month', 'day', 'hour', 'minute', 'second'
    :raise ValueError: if the string cannot be parsed as a date/time delta
    """
    from datetime import timedelta
    dt_string = dt_string.upper()

    match = re.match(r'(?:[P])(?P<date>[^T :]+)?(?:[T :])?(?P<time>.+)?$', dt_string)

    if match is None:
        raise ValueError('Date/Time step must be in ISO 8601 format, for example PY2M3DT4H5M6S.')

    if match.group('date') is not None:
        date_string = match.group('date')
    else:
        date_string = ''

    if match.group('time') is not None:
        time_string = match.group('time')
    else:
        time_string = ''

    date_tokens = re.findall('[0-9]*[A-Z]', date_string)
    time_tokens = re.findall('[0-9]*[A-Z]', time_string)

    years = 0
    months = 0
    days = 0
    hours = 0
    minutes = 0
    seconds = 0

    for token in date_tokens:
        val = int(token[:-1])
        if token[-1:] == "Y":
            years = val
        elif token[-1:] == "M":
            months = val
        elif token[-1:] == "D":
            days = val
        else:
            raise ValueError("Date/Time step must be in ISO 8601 format, for example PY2M3DT4H5M6S")

    for token in time_tokens:
        val = int(token[:-1])
        if token[-1:] == "H":
            hours = val
        elif token[-1:] == "M":
            minutes = val
        elif token[-1:] == "S":
            seconds = val
        else:
            raise ValueError("Date/Time step must be in ISO 8601 format, for example PY2M3DT4H5M6S")

    # Note that there is a loss of precision here because we have to convert months and years to integer days
    dt = timedelta(seconds=seconds, minutes=minutes, hours=hours, days=(days + months*30 + years*365))
    return dt


def _datetime_delta_to_float_days(td):
    """
    Converts a timedelta into a fractional day
    :param td: the timedelta to be converted
    :return: a float representation of a day
    """
    sec = 1.0/(24.0*60.0*60.0)  # Conversion from sec to day
    return td.total_seconds()*sec


def parse_datetimestr_delta_to_float_days(string):
    """
    Parses "PY2M3DT4H5M6S" (ISO 8601) into a fractional day
    :param string: string to be parsed
    :return: a float representation of a day
    """

    date_delta = _parse_datetime_delta(string)
    return _datetime_delta_to_float_days(date_delta)


def parse_as_number_or_datetime(string):
    """
    Parse a string as a number from the command line, or if that fails, as a datetime

    :param in_string: String to parse
    :return: int, float or DateTime
    """
    if string == 'None' or string is None:
        return None
    try:
        ret = int(string)
    except ValueError:
        try:
            ret = float(string)
        except ValueError:
            try:
                ret = _parse_datetime(string)
            except ValueError:
                raise argparse.ArgumentTypeError("'{}' is not a valid value.".format(string))
    return ret


def parse_as_number_or_standard_time(string):
    """
    Parse a string as a number from the command line, or if that fails, as a datetime in standard cis units

    :param in_string: String to parse
    :return: int, float (possibly representing a time in CIS standard time units)
    """
    from datetime import datetime
    res = parse_as_number_or_datetime(string)
    if isinstance(res, datetime):
        res = cis_standard_time_unit.date2num(res)
    return res


def parse_as_number_or_datetime_delta(string):
    """
    Parse a string as a number from the command line, or if that fails, as a datetime delta

    :param in_string: String to parse
    :return: int, float value or datetimedelta
    """
    try:
        ret = int(string)
    except ValueError:
        try:
            ret = float(string)
        except ValueError:
            try:
                ret = _parse_datetime_delta(string)
            except ValueError:
                raise argparse.ArgumentTypeError("'{}' is not a valid value.".format(string))
    return ret





from operator import mul
import numpy

from cis import __version__
from functools import reduce


class EvaluationError(Exception):
    """
    Represents an exception which occurred in a CIS evaluation, possibly resulting from incorrect user input.
    """


class Calculator(object):
    """
    Class to perform arithmetic calculations on sets of data.
    """

    SAFE_BUILTINS = ['abs', 'all', 'any', 'bool', 'divmod', 'enumerate', 'filter', 'int', 'len', 'map', 'max',
                     'min', 'pow', 'range', 'reversed', 'round', 'sorted', 'sum', 'zip']
    SAFE_MODULES = ['numpy']

    def evaluate(self, data_list, expr, output_var=None, units=None, attributes=None):
        """
        Evaluate a given expression over a list of data to produce an output data
        :param data_list: List of data used in the evaluation (these will be identified by alias or var_name).
        Should all be of the same data type and shape.
        :param expr: String python expression to evaluate.
        :param output_var: Name to use for the output variable (defaults to 'calculated_variable' if not provided.
        :param units: String indicating the units of the output expression
        :param attributes: Dictionary of attribute names : values to add to the output NetCDF variable
        :return: Data object matching the type of the input data (i.e. GriddedData or UngriddedData).
        """
        import six
        if '__' in expr:
            raise EvaluationError("Use of functions or variables with double underscores (__) is not allowed")
        # Create list of allowed globals
        safe_globals = {module: globals()[module] for module in self.SAFE_MODULES}
        # Add allowed modules (should already be imported into current namespace)
        safe_globals['__builtins__'] = {var: globals()['__builtins__'][var] for var in self.SAFE_BUILTINS}
        safe_locals = {}
        for var in data_list:
            assert isinstance(var.alias, six.string_types)
            assert isinstance(var.data, numpy.ndarray)
            if not var.is_gridded:
                safe_locals[var.alias] = var.data_flattened
            else:
                safe_locals[var.alias] = var.data
        try:
            result = eval(expr, safe_globals, safe_locals)
        except NameError as ex:
            raise EvaluationError("A variable or function referenced in your expression could not be found - "
                                  "check your expression. Error is: %s" % ex.args[0])
        except ValueError as ex:
            raise EvaluationError("An error occurred evaluating your expression - check that it's correct and that "
                                  "the variables are compatible shapes. Error is: %s" % ex.args[0])
        return self._post_process(data_list, result, expr, output_var, units, attributes)

    def _post_process(self, data_list, result_array, expr, output_var, units, attributes):
        """
        Take sample data and the resultant output array, combine them and add appropriate metadata.
        :param data_list: The original sample data
        :param result_array: The calculated output array
        :param expr: The expression used to calculate the output
        :param output_var: Name to use for the output variable.
        :param units: Units of output variable
        :param attributes: Dictionary of attributes to add to output variable
        :return: Post processed data object (GriddedData, UngriddedData or Cube)
        """
        if output_var is None:
            output_var = 'calculated_variable'
        standard_name = None
        long_name = 'Calculated value for expression "%s"' % expr
        history = self._make_history(data_list, expr)
        if units is None:
            units = 'unknown'

        # Use the first item from the data list as we assume that:
        # - The coordinates are all the same
        # - The shape is all the same
        sample_data = data_list[0]
        sample_shape = sample_data.data.shape
        if not sample_data.is_gridded:
            # Allow ungridded data to be flattened
            sample_shape = (reduce(mul, sample_shape),)
        if not sample_shape == result_array.shape:
            raise EvaluationError("The resulting array is not the same shape as the original data. "
                                  "Check your expression")
        data = sample_data.make_new_with_same_coordinates(data=result_array, var_name=output_var,
                                                          standard_name=standard_name, long_name=long_name,
                                                          history=history, units=units, flatten=True)
        if attributes is not None:
            data.add_attributes(attributes)
        return data

    def _make_history(self, data_list, expr):
        """
        Generate the history string
        :param data_list: List of data used in calculation
        :param expr: Expression using in calculation
        :return: History string
        """
        var_info = self._make_var_info(data_list)
        history_template = "Evaluated using CIS version {version}\n" + \
                           "Expression evaluated: '{expr}'\n" + \
                           "with variables: {var_info}."
        return history_template.format(version=__version__, expr=expr, var_info=var_info)

    def _make_var_info(self, data_list):
        """
        Get the variable info part of the history string, containing info
        about the variable names, aliases used and filenames.
        :param data_list: List of data used in calculation.
        :return: Var info string
        """
        var_infos = []
        var_info_tmpl = "'{var_name}'{alias} from files {files}"
        for var in data_list:
            alias = ''
            # If the alias is defaulting to the var_name then it probably hasn't been set
            if var.alias != var.var_name:
                alias = " (as '%s')" % var.alias
            var_info = var_info_tmpl.format(var_name=var.var_name, alias=alias, files=var.filenames)
            var_infos.append(var_info)
        return ",\n".join(var_infos)



from abc import ABCMeta, abstractmethod
import six


@six.add_metaclass(ABCMeta)
class CommonData(object):
    """
    Interface of common methods implemented for gridded and ungridded data.
    """
    filenames = []

    _alias = None

    @property
    @abstractmethod
    def history(self):
        """
        Return the associated history of the object

        :return: The history
        :rtype: str
        """
        return None

    @property
    def alias(self):
        """
        Return an alias for the variable name. This is an alternative name by which this data object may be identified
        if, for example, the actual variable name is not valid for some use (such as performing a python evaluation).

        :return: The alias
        :rtype: str
        """
        if self._alias is not None:
            return self._alias
        else:
            return self.var_name

    @alias.setter
    def alias(self, alias):
        """
        Set this data objects alias - this is an alternative name by which this data object may be identified
        if, for example, the actual variable name is not valid for some use (such as performing a python evaluation).

        :param str alias: The alias to use
        """
        self._alias = alias

    @property
    @abstractmethod
    def var_name(self):
        """
        Return the variable name associated with this data object

        :return: The variable name
        """
        return None

    @abstractmethod
    def get_coordinates_points(self):
        """Returns a list-like object allowing access to the coordinates of all points as HyperPoints.
        The object should allow iteration over points and access to individual points.

        :return: list-like object of data points
        """
        pass

    @abstractmethod
    def get_all_points(self):
        """Returns a list-like object allowing access to all points as HyperPoints.
        The object should allow iteration over points and access to individual points.

        :return: list-like object of data points
        """
        pass

    @abstractmethod
    def get_non_masked_points(self):
        """Returns a list-like object allowing access to all points as HyperPoints.
        The object should allow iteration over non-masked points and access to individual points.

        :return: list-like object of data points
        """
        pass

    @abstractmethod
    def is_gridded(self):
        """Returns value indicating whether the data/coordinates are gridded.
        """
        pass

    @abstractmethod
    def as_data_frame(self, copy):
        """
        Convert a CommonData object to a Pandas DataFrame.

        :param copy: Create a copy of the data for the new DataFrame? Default is True.
        :return: A Pandas DataFrame representing the data and coordinates. Note that this won't include any metadata.
        """
        pass

    @abstractmethod
    def set_longitude_range(self, range_start):
        """
        Rotates the longitude coordinate array and changes its values by
        360 as necessary to force the values to be within a 360 range starting
        at the specified value.
        :param range_start: starting value of required longitude range
        """
        pass

    @abstractmethod
    def subset(self, **kwargs):
        """
        Subset the CommonData object based on the specified constraints. Constraints on arbitrary coordinates are
        specified using keyword arguments. Each constraint must have two entries (a maximum and a minimum) although
        one of these can be None. Datetime objects can be used to specify upper and lower datetime limits, or a
        single PartialDateTime object can be used to specify a datetime range.

        The keyword keys are used to find the relevant coordinate, they are looked for in order of name, standard_name,
        axis and var_name.

        For example:
            data.subset(time=[datetime.datetime(1984, 8, 28), datetime.datetime(1984, 8, 29)],
                             altitude=[45.0, 75.0])

        Will subset the data from the start of the 28th of August 1984, to the end of the 29th, and between altitudes of
        45 and 75 (in whatever units ares used for that Coordinate).

        And:
            data.subset(time=[PartialDateTime(1984, 9)])

        Will subset the data to all of September 1984.

        :param kwargs: The constraint arguments
        :return CommonData: The subset of the data
        """
        pass

    @abstractmethod
    def sampled_from(self, data, how='', kernel=None, missing_data_for_missing_sample=True, fill_value=None,
                     var_name='', var_long_name='', var_units='', **kwargs):
        """
        Collocate the CommonData object with another CommonData object using the specified collocator and kernel

        :param CommonData or CommonDataList data: The data to resample
        :param str how: Collocation method (e.g. lin, nn, bin or box)
        :param str or cis.collocation.col_framework.Kernel kernel:
        :param bool missing_data_for_missing_sample: Should missing values in sample data be ignored for collocation?
        :param float fill_value: Value to use for missing data
        :param str var_name: The output variable name
        :param str var_long_name: The output variable's long name
        :param str var_units: The output variable's units
        :param kwargs: Constraint arguments such as h_sep, a_sep, etc.
        :return CommonData: The collocated dataset
        """
        pass

    def collocated_onto(self, sample, how='', kernel=None, missing_data_for_missing_sample=True, fill_value=None,
                        var_name='', var_long_name='', var_units='', **kwargs):
        """
        Collocate the CommonData object with another CommonData object using the specified collocator and kernel.

        :param CommonData sample: The sample data to collocate onto
        :param str how: Collocation method (e.g. lin, nn, bin or box)
        :param str or cis.collocation.col_framework.Kernel kernel:
        :param bool missing_data_for_missing_sample: Should missing values in sample data be ignored for collocation?
        :param float fill_value: Value to use for missing data
        :param str var_name: The output variable name
        :param str var_long_name: The output variable's long name
        :param str var_units: The output variable's units
        :param kwargs: Constraint arguments such as h_sep, a_sep, etc.
        :return CommonData: The collocated dataset
        """
        return sample.sampled_from(self, how=how, kernel=kernel,
                                   missing_data_for_missing_sample=missing_data_for_missing_sample,
                                   fill_value=fill_value, var_name=var_name, var_long_name=var_long_name,
                                   var_units=var_units, **kwargs)

    def plot(self, *args, **kwargs):
        """
        Plot the data. A matplotlib Axes is created if none is provided.

        The default method for series data is 'line', otherwise (for e.g. a map plot) is 'scatter2d' for UngriddedData
        and 'heatmap' for GriddedData.

        :param string how: The method to use, one of:  "contour", "contourf", "heatmap", "line", "scatter", "scatter2d",
        "comparativescatter", "histogram", "histogram2d" or "taylor"
        :param Axes ax: A matplotlib axes on which to draw the plot
        :param Coord or CommonData xaxis: The data to plot on the x axis
        :param Coord or CommonData yaxis: The data to plot on the y axis
        :param string or cartopy.crs.Projection projection: The projection to use for map plots (default is PlateCaree)
        :param float central_longitude: The central longitude to use for PlateCaree (if no other projection specified)
        :param string label: A label for the data. This is used for the title, colorbar or legend depending on plot type
        :param args: Other plot-specific args
        :param kwargs: Other plot-specific kwargs
        :return Axes: The matplotlib Axes on which the plot was drawn
        """
        from cis.plotting.plot import basic_plot
        _, ax = basic_plot(self, *args, **kwargs)
        return ax

    @abstractmethod
    def _get_default_plot_type(self, lat_lon=False):
        pass

    def _get_coord(self, name):
        from cis.utils import standard_axes
        def _try_coord(data, coord_dict):
            import cis.exceptions as cis_ex
            import iris.exceptions as iris_ex
            try:
                coord = data.coord(**coord_dict)
            except (iris_ex.CoordinateNotFoundError, cis_ex.CoordinateNotFoundError):
                coord = None
            return coord

        coord = _try_coord(self, dict(name_or_coord=name)) or _try_coord(self, dict(standard_name=name)) \
            or _try_coord(self, dict(standard_name=standard_axes.get(name.upper(), None))) or \
                _try_coord(self, dict(var_name=name)) or _try_coord(self, dict(axis=name))

        return coord


@six.add_metaclass(ABCMeta)
class CommonDataList(list):
    """
    Interface for common list methods implemented for both gridded and ungridded data.

    Note that all objects in a CommonDataList must have the same coordinates and coordinate values.
    """

    def __init__(self, iterable=()):
        super(CommonDataList, self).__init__()
        self.extend(iterable)

    def __add__(self, rhs):
        return self.__class__(list.__add__(self, rhs))

    def __getitem__(self, item):
        result = list.__getitem__(self, item)
        if isinstance(result, list):
            result = self.__class__(result)
        return result

    def __getslice__(self, start, stop):
        result = super(CommonDataList, self).__getslice__(start, stop)
        result = self.__class__(result)
        return result

    def __str__(self):
        """Runs short :meth:`CommonData.summary` on every item and the coords"""
        result = ['%s: %s' % (i, item.summary(shorten=True)) for i, item in enumerate(self)]
        if result:
            result = ''.join(result)
            result += 'Coordinates: \n'
            for c in self.coords():
                result += '{pad:{width}}{name}\n'.format(pad=' ', width=2, name=c.name())
        else:
            result = '< Empty list >'
        return result

    def __repr__(self):
        """Runs repr on every cube."""
        return '[%s]' % ',\n'.join([repr(item) for item in self])

    @property
    @abstractmethod
    def is_gridded(self):
        """
        Returns value indicating whether the data/coordinates are gridded.
        """
        raise NotImplementedError

    def append(self, p_object):
        """
        Append a compatible object to this list

        :param UngriddedData or GriddedData p_object:
        :return: None
        :raises TypeError: If the objects being added are of the wrong type (i.e. Gridded versus Ungridded).
        :raises ValueError: If the coordinates aren't compatible with each other.
        """
        this_class = self.__class__.__name__
        that_class = p_object.__class__.__name__
        other_gridded = getattr(p_object, 'is_gridded', None)

        if other_gridded is None or other_gridded != self.is_gridded:
            raise TypeError("Appending {that_class} to {this_class} is not allowed".format(
                that_class=that_class, this_class=this_class))

        if len(self) > 0:
            if len(self.coords()) != len(p_object.coords()):
                raise ValueError("Incompatible data: {} must have the same number of coordinates as the first element "
                                 "in this list.".format(that_class))
            for this_coord, that_coord in zip(self.coords(), p_object.coords()):
                if any([(a != b) for a, b in zip(this_coord.points.shape, that_coord.points.shape)]):
                    raise ValueError("Given coordinate {} has shape {} which is "
                                     "incompatible with {}.".format(that_coord.name(),
                                                                    that_coord.shape,
                                                                    this_coord.shape))

        super(CommonDataList, self).append(p_object)
        return

    def extend(self, iterable):
        # This will raise a TypeError if the objects being added are of the wrong type (i.e. Gridded versus Ungridded).
        for x in iterable:
            self.append(x)

    def append_or_extend(self, item_to_add):
        """
        Append or extend an item to an existing list, depending on whether the item to add is itself a list or not.
        :param item_to_add: Item to add (may be list or not).
        """
        if isinstance(item_to_add, list):
            self.extend(item_to_add)
        else:
            self.append(item_to_add)

    @property
    def var_name(self):
        """
        Get the variable names in this list
        """
        var_names = []
        for data in self:
            var_names.append(data.var_name)
        return var_names

    @property
    def filenames(self):
        """
        Get the filenames in this data list
        """
        filenames = []
        for data in self:
            filenames.extend(data.filenames)
        return filenames

    @filenames.setter
    def filenames(self, names):
        """
        Add these filenames to this data list
        """
        for data in self:
            data.filenames.extend(names)

    def add_history(self, new_history):
        """
        Appends to, or creates, the metadata history attribute using the supplied history string.
        The new entry is prefixed with a timestamp.
        :param new_history: history string
        """
        for data in self:
            data.add_history(new_history)

    def coords(self, *args, **kwargs):
        """
        Returns all coordinates used in all the data object
        :return: A list of coordinates in this data list object fitting the given criteria
        """
        return self[0].coords(*args, **kwargs)

    def set_longitude_range(self, range_start):
        """
        Rotates the longitude coordinate array and changes its values by
        360 as necessary to force the values to be within a 360 range starting
        at the specified value.
        :param range_start: starting value of required longitude range
        """
        for data in self:
            data.set_longitude_range(range_start)

    @abstractmethod
    def subset(self, **kwargs):
        """
        Subset the CommonDataList object based on the specified constraints. Constraints on arbitrary coordinates are
        specified using keyword arguments. Each constraint must have two entries (a maximum and a minimum) although
        one of these can be None. Datetime objects can be used to specify upper and lower datetime limits, or a
        single PartialDateTime object can be used to specify a datetime range.

        The keyword keys are used to find the relevant coordinate, they are looked for in order of name, standard_name,
        axis and var_name.

        For example:
            data.subset(time=[datetime.datetime(1984, 8, 28), datetime.datetime(1984, 8, 29)],
                             altitude=[45.0, 75.0])

        Will subset the data from the start of the 28th of August 1984, to the end of the 29th, and between altitudes of
        45 and 75 (in whatever units ares used for that Coordinate).

        And:
            data.subset(time=[PartialDateTime(1984, 9)])

        Will subset the data to all of September 1984.

        :param kwargs: The constraint arguments
        :return CommonDataList: The subset of each of the data
        """
        pass

    def collocated_onto(self, sample, how='', kernel=None, missing_data_for_missing_sample=True, fill_value=None,
                        var_name='', var_long_name='', var_units='', **kwargs):
        """
        Collocate the CommonData object with another CommonData object using the specified collocator and kernel.

        :param CommonData sample: The sample data to collocate onto
        :param str how: Collocation method (e.g. lin, nn, bin or box)
        :param str or cis.collocation.col_framework.Kernel kernel:
        :param bool missing_data_for_missing_sample: Should missing values in sample data be ignored for collocation?
        :param float fill_value: Value to use for missing data
        :param str var_name: The output variable name
        :param str var_long_name: The output variable's long name
        :param str var_units: The output variable's units
        :param kwargs: Constraint arguments such as h_sep, a_sep, etc.
        :return CommonData: The collocated dataset
        """
        return sample.sampled_from(self, how=how, kernel=kernel,
                                   missing_data_for_missing_sample=missing_data_for_missing_sample,
                                   fill_value=fill_value, var_name=var_name, var_long_name=var_long_name,
                                   var_units=var_units, **kwargs)

    def plot(self, *args, **kwargs):
        """
        Plot the data. A matplotlib Axes is created if none is provided.

        The default method for series data is 'line', otherwise (for e.g. a map plot) is 'scatter2d' for UngriddedData
        and 'heatmap' for GriddedData.

        :param string how: The method to use, one of:  "contour", "contourf", "heatmap", "line", "scatter", "scatter2d",
        "comparativescatter", "histogram", "histogram2d" or "taylor"
        :param Axes ax: A matplotlib axes on which to draw the plot
        :param Coord or CommonData xaxis: The data to plot on the x axis
        :param Coord or CommonData yaxis: The data to plot on the y axis
        :param list layer_opts: A list of keyword dictionaries to pass to each layer of the plot.
        :param args: Other plot-specific args to pass to all plots
        :param kwargs: Other plot-specific kwargs to pass to all plots
        :return Axes: The matplotlib Axes on which the plot was drawn
        """
        from cis.plotting.plot import multilayer_plot
        _, ax = multilayer_plot(self, *args, **kwargs)
        return ax

    def _get_coord(self, *args):
        return self[0]._get_coord(*args)



# Based on examples here: http://erikzaadi.com/2012/07/03/mocking-python-imports/
"""
    Tests for checking correct behaviour when Python HDF is not installed. ImportError should be raised when HDF is used
    not when CIS is first started.
"""
from nose.tools import raises
from mock import MagicMock, patch


class HDFNotInstalledTests(object):

    def setup(self):
        """
        We have to patch the sys.modules dictionary to use our mocked versions of the pyhdf library to test it not
        being installed. This creates the appropriate mocks and imports them into the namespace.
        """

        self.pyhdf_mock = MagicMock()
        self.pyhdf_mock.SD = None
        self.pyhdf_mock.HDF = None
        modules = {
            'pyhdf': self.pyhdf_mock,
            'pyhdf.SD': self.pyhdf_mock.SD,
            'pyhdf.HDF': self.pyhdf_mock.HDF
        }

        self.module_patcher = patch.dict('sys.modules', modules)
        self.module_patcher.start()

        from cis.data_io import hdf, hdf_vd, hdf_sd
        self.hdf_vd = hdf_vd
        self.hdf_sd = hdf_sd
        self.hdf = hdf

    def teardown(self):
        """
        Remove the patches
        """
        self.module_patcher.stop()

    @raises(ImportError)
    def test_no_pyhdf_raises_not_installed_in_VD_variables(self):
        _ = self.hdf_vd.get_hdf_VD_file_variables('some_file')

    @raises(ImportError)
    def test_no_pyhdf_raises_not_installed_in_VD_read(self):
        _ = self.hdf_vd.read('some_file', 'some_variable')

    @raises(ImportError)
    def test_no_pyhdf_raises_not_installed_in_SD_variables(self):
        _ = self.hdf_sd.get_hdf_SD_file_variables('some_file')

    @raises(ImportError)
    def test_no_pyhdf_raises_not_installed_in_SD_read(self):
        _ = self.hdf_sd.read('some_file', 'some_variable')

    @raises(ImportError)
    def test_no_pyhdf_raises_not_installed_in_HDF_get_metadata(self):
        _ = self.hdf.get_hdf4_file_metadata('some_file')



"""
Abstract classes defining the plotting interface for all CIS plot types. Each Plot deals with one 'layer' or set of
data.
"""
import logging
from abc import ABCMeta, abstractmethod
import six
from cis.plotting.formatted_plot import set_log_scales
from iris.coords import POINT_MODE


@six.add_metaclass(ABCMeta)
class APlot(object):

    def __init__(self, packed_data, xaxis, yaxis=None, color=None,
                 edgecolor=None, itemstyle=None, itemwidth=None, label=None, *mplargs, **mplkwargs):
        """
        Abstract base class for all CIS plot types. Includes some common kwargs.

        :param CommonData packed_data: A packed data item (as opposed to a Numpy array)
        :param Coord or CommonData xaxis: A data item for the xaxis (optional depending on plot type)
        :param Coord or CommonData yaxis: A data item for the yaxis (optional depending on plot type)
        :param string color: A color for the plot (line or scatter color for example)
        :param string edgecolor: The edgecolor to use for scatter and comparative plots
        :param string itemstyle: The itemstyle to use for scatter and comparative plots
        :param int itemwidth: The width for the plot items (lines or points)
        :param string label: A label for the plot layer - this is used either as a title or legend label
        :param mplargs: Any arguments to be passed directly into matplotlib
        :param mplkwargs: Any keyword arguments to be passed directly into matplotlib
        """
        # Raw data attributes (for unpacking the packed data into)
        self.data = None
        self.x = None
        self.y = None

        self.xaxis = xaxis
        self.yaxis = yaxis
        self.color = color
        self.label = label
        self.edgecolor = edgecolor
        self.itemstyle = itemstyle
        self.itemwidth = itemwidth

        self.mplargs = mplargs
        self.mplkwargs = mplkwargs

        self.color_axis = []

    @abstractmethod
    def __call__(self, ax):
        """
        The method that will do the plotting. To be implemented by each subclass of APlot.
        """
        pass

    @abstractmethod
    def is_map(self):
        """
        :return: A boolean saying if this plot should be represented as a map
        """
        pass

    def is_rectangular_projection(self, ax):
        """
        :return: A boolean saying if this (map) plot is on a rectangular projection
        """
        from cartopy.crs import _RectangularProjection
        return self.is_map() and isinstance(ax.projection, _RectangularProjection)

    def set_log_scales(self, ax, logx, logy):
        """
        Optionally log-scale one or both of the axis. Default is to rescale after logging.
        """
        set_log_scales(ax, logx, logy, True)


class GenericPlot(APlot):

    def __init__(self, packed_data, *args, **kwargs):
        """
        A generic (abstract) class for plotting data against a single coordinate, as e.g. a line or scatter.
        """
        from cis.plotting.plot import get_label
        super(GenericPlot, self).__init__(packed_data, *args, **kwargs)

        logging.debug("Unpacking the data items")

        if self.xaxis.standard_name == 'time':
            # Convert to matplotlib datetime
            self.x = self.xaxis.units.convert(self.xaxis.points, 'days since 1970-01-01 00:00:00')
        else:
            self.x = self.xaxis.points

        self.data = packed_data.data

        self.mplkwargs['label'] = self.label or packed_data.name()
        self.xlabel, self.ylabel = get_label(self.xaxis), get_label(packed_data)

    def __call__(self, ax):
        ax.set_xlabel(self.xlabel)
        ax.set_ylabel(self.ylabel)

    def is_map(self):
        return False


class Generic2DPlot(APlot):

    # The mode is used for unpacking gridded data objects to define how to unpack the coordinates
    MODE = POINT_MODE

    def __init__(self, packed_data, logv=None, vstep=None,
                 colourbar=True, cbarscale=None, cbarorient=None, cbarlabel=None,
                 coastlines=True, coastlinescolour='k', *args, **kwargs):
        """
        A generic (abstract) class for plotting data against two coordinates, as e.g. a heatmap or contour. This
        includes a number of kwargs common to this type of plot.

        :param CommonData packed_data: The data to plot on the v (colour) axis
        :param bool logv: Log the v (colour bar) axis?
        :param float vstep: The step to use for the v axis (colour bar)
        :param bool colourbar: Include colour bar? Default True
        :param float cbarscale: A scale factor for the colorbar
        :param string cbarorient: The colour bar orientation ('vertical' or 'horizontal')
        :param string cbarlabel: A label for the colour bar
        :param bool coastlines: Plot coastlines on map plots? Default True
        :param string coastlinescolour: The colour to use for coastlines (default black)
        """
        from .plot import get_label
        super(Generic2DPlot, self).__init__(packed_data, *args, **kwargs)

        logging.debug("Unpacking the data items")
        self.data, self.x, self.y = self.unpack_data_items(packed_data)

        self.xlabel = get_label(self.xaxis)
        self.ylabel = get_label(self.yaxis)

        self.logv = logv
        self.vstep = vstep
        self.cbarscale = cbarscale
        if cbarorient is None:
            # Set the color bar to be horizontal if we're plotting a wide map, otherwise vertical
            self.cbarorient = 'horizontal' if self.is_map() and \
                                              (self.x.max() - self.x.min()) > \
                                              (self.y.max() - self.y.min()) else 'vertical'
        else:
            self.cbarorient = cbarorient

        self.colourbar = colourbar
        self.cbarlabel = cbarlabel or self.label

        self.coastlines = coastlines
        self.coastlinescolour = coastlinescolour

        if self.logv:
            from matplotlib.colors import LogNorm
            self.mplkwargs["norm"] = LogNorm()

    def __call__(self, ax):
        from .plot import add_color_bar
        from .plot import drawcoastlines, get_best_map_ticks, set_map_ticks
        ax.set_xlabel(self.xlabel)
        ax.set_ylabel(self.ylabel)

        # Either setup the colour bar - or if we're not doing one then we need to set the title
        if self.colourbar:
            add_color_bar(ax, self.mappable, self.vstep, self.logv, self.cbarscale, self.cbarorient, self.cbarlabel)
        else:
            ax.set_title(self.label)

        if self.is_map():
            if self.coastlines:
                drawcoastlines(ax, self.coastlinescolour)
            set_map_ticks(ax, *get_best_map_ticks(ax))

    def is_map(self):
        if self.xaxis.name().lower().startswith("lon") and self.yaxis.name().lower().startswith("lat"):
            return True

    def set_log_scales(self, ax, logx, logy):
        if self.is_map():
            logging.debug("Ignoring log formatting for map plot")
        else:
            # Don't rescale as it doesn't work for collections (heatmaps, scatter2d etc)
            set_log_scales(ax, logx, logy, False)

    def unpack_data_items(self, data_object):
        """
        :param CommonData data_object: A cube or an UngriddedData object
        :return: data, x, and y and as numpy arrays
        """
        from iris.cube import Cube
        import logging

        if isinstance(data_object, Cube):
            x = self.xaxis.points if self.MODE == POINT_MODE else self.xaxis.contiguous_bounds()
            y = self.yaxis.points if self.MODE == POINT_MODE else self.yaxis.contiguous_bounds()
            # We need to do some more unpacking, and possibly meshgridding for gridded data
            data, x, y = self._cube_manipulation(data_object, x, y)
        else:
            x = self.xaxis.points
            y = self.yaxis.points
            data = data_object.data

        logging.debug("Shape of x: " + str(x.shape))
        if y is not None:
            logging.debug("Shape of y: " + str(y.shape))
        logging.debug("Shape of data: " + str(data.shape))

        return data, x, y

    @staticmethod
    def _cube_manipulation(cube, x, y):
        """
        Optionally transpose the data and make a mesh-grid, taking into account auxilliary coordinates
        A lot of this logic closely follows that used in iris.qplt.

        :param Cube cube: An iris Cube
        :param NDarray x: a numpy array of x points
        :param NDarray y: a numpy array of y points
        :return: data, x, and y and as numpy arrays
        """
        import iris.plot as iplt
        import iris
        from cartopy.util import add_cyclic_point
        import numpy as np

        no_of_dims = len(cube.shape)
        data = cube.data

        plot_defn = iplt._get_plot_defn(cube, iris.coords.POINT_MODE, ndims=no_of_dims)
        if plot_defn.transpose:
            data = data.T
            x = x.T
            y = y.T

        # Check for auxiliary coordinates.
        aux_coords = False
        for coord in cube[0].coords(dim_coords=False):
            aux_coords = True
        if no_of_dims == 2:
            # If we have found some auxiliary coordinates in the data and the shape of x data or y data is the same as
            # data assume we have a hybrid coordinate (which is two dimensional b nature. Perhaps need a more robust
            # method for detecting this.
            if aux_coords and (data.shape == x.shape or data.shape == y.shape):
                # Work out which set of data needs expanding to match the coordinates of the others. Note there can only
                # ever be one hybrid coordinate axis.
                if y.shape == data.shape:
                    if y[:, 0].shape == x.shape:
                        x, _y = np.meshgrid(x, y[0, :])
                    elif y[0, :].shape == x.shape:
                        x, _y = np.meshgrid(x, y[:, 0])
                elif x.shape == data.shape:
                    if x[:, 0].shape == y.shape:
                        y, _x = np.meshgrid(y, x[0, :])
                    elif x[0, :].shape == y.shape:
                        y, _x = np.meshgrid(y, x[:, 0])
            else:
                if len(x) == data.shape[-1]:
                    try:
                        data, x = add_cyclic_point(data, x)
                    except ValueError as e:
                        logging.warn('Unable to add cyclic data point for x-axis. Error was: ' + e.args[0])
                    x, y = np.meshgrid(x, y)
                elif len(y) == data.shape[-1]:
                    try:
                        data, y = add_cyclic_point(data, y)
                    except ValueError as e:
                        logging.warn('Unable to add cyclic data point for y-axis. Error was: ' + e.args[0])
                    y, x = np.meshgrid(y, x)
        return data, x, y



import abc
import six
import numpy
import scipy.stats.mstats
from iris.cube import Cube


@six.add_metaclass(abc.ABCMeta)
class StatisticsResult(object):
    """
    Holds statistical calculations performed by a statistics analysis on two files.
    """

    @abc.abstractmethod
    def pprint(self):
        """
        Nicely formatted string representation of this statistical result, suitable for printing to screen.
        """

    @abc.abstractmethod
    def as_cube(self):
        """
        Get this statistical result as an iris.cube.Cube instance
        """


class PointsCount(StatisticsResult):
    """
    Number of points used in statistical calculation (masked or missing points excluded).
    """

    def __init__(self, num_points):
        self.num_points = num_points

    def pprint(self):
        """
        Nicely formatted string representation of this statistical result, suitable for printing to screen.
        """
        return "Number of points: %s" % self.num_points

    def as_cube(self):
        """
        Get this statistical result as an iris.cube.Cube instance
        """
        cube = Cube([self.num_points], long_name='Number of points used in calculations',  var_name='num_points')
        return cube


class DatasetMean(StatisticsResult):
    """
    Mean of an individual dataset
    """

    def __init__(self, mean, ds_name, ds_no):
        self.mean = mean
        self.ds_name = ds_name
        self.ds_no = ds_no

    def pprint(self):
        """
        Nicely formatted string representation of this statistical result, suitable for printing to screen.
        """
        return "Mean value of dataset %s: %s" % (self.ds_no, self.mean)

    def as_cube(self):
        """
        Get this statistical result as an iris.cube.Cube instance
        """
        return Cube(self.mean, long_name='Mean value of %s' % self.ds_name, var_name='dataset_mean_%s' % self.ds_no)


class DatasetStddev(StatisticsResult):
    """
    Standard deviations of an individual dataset
    """

    def __init__(self, stddev, ds_name, ds_no):
        self.stddev = stddev
        self.ds_name = ds_name
        self.ds_no = ds_no

    def pprint(self):
        """
        Nicely formatted string representation of this statistical result, suitable for printing to screen.
        """
        return "Standard deviation for dataset %s: %s" % (self.ds_no, self.stddev)

    def as_cube(self):
        """
        Get this statistical result as an iris.cube.Cube instance
        """
        return Cube(self.stddev, long_name='Corrected sample standard deviation of %s' % self.ds_name,
                    var_name='dataset_stddev_%s' % self.ds_no)


class AbsoluteMean(StatisticsResult):
    """
    Mean of the absolute difference between the two datasets (i.e, mean(data2 - data1))
    """

    def __init__(self, abs_mean):
        self.abs_mean = abs_mean

    def pprint(self):
        """
        Nicely formatted string representation of this statistical result, suitable for printing to screen.
        """
        return "Mean of absolute difference: %s" % self.abs_mean

    def as_cube(self):
        """
        Get this statistical result as an iris.cube.Cube instance
        """
        cube = Cube([self.abs_mean], long_name='Mean of the absolute difference (data2 - data1)', var_name='abs_mean')
        return cube


class AbsoluteStddev(StatisticsResult):
    """
    Standard deviation of the absolute difference between the two datasets (i.e, mean(data2 - data1))
    """

    def __init__(self, abs_stddev):
        self.abs_stddev = abs_stddev

    def pprint(self):
        """
        Nicely formatted string representation of this statistical result, suitable for printing to screen.
        """
        return "Standard deviation of absolute difference: %s" % self.abs_stddev

    def as_cube(self):
        """
        Get this statistical result as an iris.cube.Cube instance
        """
        cube = Cube([self.abs_stddev], var_name='abs_stddev',
                    long_name='Corrected sample standard deviation of the absolute difference (data2 - data1)')
        return cube


class RelativeMean(StatisticsResult):
    """
    Mean of the relative difference between the two datasets (i.e, mean((data2 - data1)/data1))
    """

    def __init__(self, rel_mean):
        self.rel_mean = rel_mean

    def pprint(self):
        """
        Nicely formatted string representation of this statistical result, suitable for printing to screen.
        """
        return "Mean of relative difference: %s" % self.rel_mean

    def as_cube(self):
        """
        Get this statistical result as an iris.cube.Cube instance
        """
        cube = Cube([self.rel_mean], long_name='Mean of the relative difference (data2 - data1)/data1',
                    var_name='rel_mean')
        return cube


class RelativeStddev(StatisticsResult):
    """
    Standard deviation of the relative difference between the two datasets (i.e, mean((data2 - data1)/data1))
    """

    def __init__(self, rel_stddev):
        self.rel_stddev = rel_stddev

    def pprint(self):
        """
        Nicely formatted string representation of this statistical result, suitable for printing to screen.
        """
        return "Standard deviation of relative difference: %s" % self.rel_stddev

    def as_cube(self):
        """
        Get this statistical result as an iris.cube.Cube instance
        """
        cube = Cube([self.rel_stddev], var_name='rel_stddev',
                    long_name='Corrected sample standard deviation of the relative difference (data2 - data1)/data1')
        return cube


class SpearmansRank(StatisticsResult):
    """
    Spearman's rank correlation coefficient
    """

    def __init__(self, spearman):
        self.spearman = spearman

    def pprint(self):
        """
        Nicely formatted string representation of this statistical result, suitable for printing to screen.
        """
        return "Spearman's rank coefficient: %s" % self.spearman

    def as_cube(self):
        """
        Get this statistical result as an iris.cube.Cube instance
        """
        cube = Cube([self.spearman], var_name='spearman', long_name="Spearman's rank correlation coefficient")
        return cube


class LinearRegressionGradient(StatisticsResult):
    """
    Linear regression gradient
    """

    def __init__(self, grad):
        self.grad = grad

    def pprint(self):
        """
        Nicely formatted string representation of this statistical result, suitable for printing to screen.
        """
        return "Linear regression gradient: %s" % self.grad

    def as_cube(self):
        """
        Get this statistical result as an iris.cube.Cube instance
        """
        return Cube(self.grad, long_name='Linear regression gradient', var_name='regression_gradient')


class LinearRegressionIntercept(StatisticsResult):
    """
    Linear regression intercept
    """

    def __init__(self, intercept):
        self.intercept = intercept

    def pprint(self):
        """
        Nicely formatted string representation of this statistical result, suitable for printing to screen.
        """
        return "Linear regression intercept: %s" % self.intercept

    def as_cube(self):
        """
        Get this statistical result as an iris.cube.Cube instance
        """
        return Cube(self.intercept, long_name='Linear regression intercept', var_name='regression_intercept')


class LinearRegressionRValue(StatisticsResult):
    """
    Linear regression r-value (PMCC)
    """

    def __init__(self, r):
        self.r = r

    def pprint(self):
        """
        Nicely formatted string representation of this statistical result, suitable for printing to screen.
        """
        return "Linear regression r-value: %s" % self.r

    def as_cube(self):
        """
        Get this statistical result as an iris.cube.Cube instance
        """
        return Cube(self.r, long_name='Linear regression r-value (Pearson product-moment correlation coefficient)',
                    var_name='regression_r')


class LinearRegressionStderr(StatisticsResult):
    """
    Linear regression standard error of the estimate
    """

    def __init__(self, stderr):
        self.stderr = stderr

    def pprint(self):
        """
        Nicely formatted string representation of this statistical result, suitable for printing to screen.
        """
        return "Linear regression standard error: %s" % self.stderr

    def as_cube(self):
        """
        Get this statistical result as an iris.cube.Cube instance
        """
        return Cube(self.stderr, long_name='Linear regression standard error of the estimate',
                    var_name='regression_stderr')


class StatsAnalyzer(object):
    """
    Analyse datasets to produce statistics.
    """

    def __init__(self, data1, data2):
        """
        Create a statistics analyser for two data sets

        :param CommonData data1: First data object
        :param CommonData data2: Second data object
        """
        self._var_name_1 = data1.var_name
        self._var_name_2 = data2.var_name
        # Flatten the data, so that data of different shapes but same overall size can be compared (e.g. the case where
        # you have collocated some data).
        self._data1 = data1.data.flatten()
        self._data2 = data2.data.flatten()
        # Keep a record of whether we have already combined the masks so that we don't unnecessarily repeat it
        self._masks_combined = False

    def analyze(self):
        """
        Perform a statistical analysis on two data sets.

        :return: List of StatisticsResult instances.
        """
        out = []
        out.extend(self.points_count())
        out.extend(self.means())
        out.extend(self.stddevs())
        out.extend(self.abs_mean())
        out.extend(self.abs_stddev())
        out.extend(self.rel_mean())
        out.extend(self.rel_stddev())
        out.extend(self.spearmans_rank())
        out.extend(self.linear_regression())
        return out

    def points_count(self):
        """
        Count all points which will be used for statistical comparison operations
        (i.e. are non-missing in both datasets).

        :return: List of StatisticsResults
        """
        count = numpy.ma.count(self._data1 + self._data2)
        return [PointsCount(count)]

    def means(self):
        """
        Means of two datasets

        :return: List of StatisticsResults
        """
        from cis.utils import apply_intersection_mask_to_two_arrays
        # This command requires the masks to be combined
        data1_combined, data2_combined = apply_intersection_mask_to_two_arrays(self._data1, self._data2)
        return [DatasetMean(numpy.mean(data1_combined), self._var_name_1, 1),
                DatasetMean(numpy.mean(data2_combined), self._var_name_2, 2)]

    def stddevs(self):
        """
        Corrected sample standard deviation of datasets

        :return: List of StatisticsResults
        """
        from cis.utils import apply_intersection_mask_to_two_arrays
        # This command requires the masks to be combined
        data1_combined, data2_combined = apply_intersection_mask_to_two_arrays(self._data1, self._data2)
        return [DatasetStddev(numpy.std(data1_combined, ddof=1), self._var_name_1, 1),
                DatasetStddev(numpy.std(data2_combined, ddof=1), self._var_name_2, 2)]

    def abs_mean(self):
        """
        Mean of absolute difference d2-d1

        :return: List of StatisticsResults
        """
        mean = numpy.mean(self._data2 - self._data1)
        return [AbsoluteMean(mean)]

    def abs_stddev(self):
        """
        Standard deviation of absolute difference d2-d1

        :return: List of StatisticsResults
        """
        stddev = numpy.std(self._data2 - self._data1, ddof=1)
        return [AbsoluteStddev(stddev)]

    def rel_mean(self):
        """
        Mean of relative difference (d2-d1)/d1

        :return: List of StatisticsResults
        """
        mean = numpy.mean((self._data2 - self._data1)/self._data1)
        return [RelativeMean(mean)]

    def rel_stddev(self):
        """
        Mean of relative difference (d2-d1)/d1

        :return: List of StatisticsResults
        """
        stddev = numpy.std((self._data2 - self._data1)/self._data1, ddof=1)
        return [RelativeStddev(stddev)]

    def spearmans_rank(self):
        """
        Perform a spearman's rank on the data

        :return: List of StatisticsResults
        """
        spearman = scipy.stats.mstats.spearmanr(self._data1, self._data2)[0]
        return [SpearmansRank(spearman)]

    def linear_regression(self):
        """
        Perform a linear regression on the data

        :return: List of StatisticsResults
        """
        grad, intercept, r, p, stderr = scipy.stats.mstats.linregress(self._data1, self._data2)
        return [LinearRegressionGradient(grad),
                LinearRegressionIntercept(intercept),
                LinearRegressionRValue(r),
                LinearRegressionStderr(stderr)]



import collections.abc
import re
import logging
import warnings
import numpy as np
from cis.exceptions import InvalidCommandLineOptionError
import contextlib


# number of bytes in a MB
BYTES_IN_A_MB = 1048576.0


def add_element_to_list_in_dict(my_dict, key, value):
    try:
        my_dict[key].append(value)
    except KeyError:
        my_dict[key] = [value]


def concatenate(arrays, axis=0):
    """
    Concatenate a list of numpy arrays into one larger array along the axis specified (the default axis is zero). If any
    of the arrays are masked arrays then the returned array will be a masked array with the correct mask, otherwise a
    numpy array is returned.

    :param arrays: A list of numpy arrays (masked or not)
    :param axis: The axis along which to concatenate (the default is 0)
    :return: The concatenated array
    """
    from numpy.ma import MaskedArray
    if any(isinstance(array, MaskedArray) for array in arrays):
        from numpy.ma import concatenate
    else:
        from numpy import concatenate

    res = arrays[0]

    if len(arrays) > 1:
        for array in arrays[1:]:
            res = concatenate((res, array), axis)

    return res


def calculate_histogram_bin_edges(data, axis, user_min, user_max, step, log_scale=False):
    """
    :param data: A numpy array
    :param axis: The axis on which the data will be plotted. Set to "x" for histogram2d
    :param user_range: A dictionary containing the min and max values for the edges specified by the user. The data min
     and max is used if the user did not specify
    :param step: The distance between each bin edge/the width of each bin
    :return: An array containing a list of bin edges (i.e. when each bin starts and ends)
    """
    from decimal import Decimal
    from numpy import array, append, logspace, log10
    import logging
    import sys

    min_val = user_min or data.min()
    max_val = user_max or data.max()

    val_range = float(max_val - min_val)

    if step is None:
        step = val_range / 10.0
        user_specified_step = False
    else:
        user_specified_step = True

    x = float(Decimal(str(val_range)) % Decimal(str(step)))

    stop = max_val + step if x < 1.e-7 else max_val  # 1.e-7 is very close to 0

    bin_edges = array([])
    i = min_val
    if not log_scale or user_specified_step:
        while abs(i - stop) >= sys.float_info.min and i < stop:
            if not user_specified_step and len(bin_edges) == 11:
                break
            bin_edges = append(bin_edges, i)

            i += step
    else:
        if min_val < 0 or max_val < 0:
            raise ValueError('Error, log axis specified but data minimum is less than 0. Please either use a linear'
                             'axis or specify an axis minimum greater than 0.')
        bin_edges = logspace(log10(min_val), log10(stop), num=11)

    logging.debug(axis + " axis bin edges: " + str(bin_edges))
    return bin_edges


def expand_1d_to_2d_array(array, length, axis=0):
    """
    General utility routine to 'extend arbitrary dimensional array into a higher dimension
    by duplicating the data along a given 'axis' (default is 0) of size 'length'.

    Examples::

        >>> a = np.array([1, 2, 3, 4])
        >>> expand_1d_to_2d_array(a, 4, axis=0)
        [[1 2 3 4]
         [1 2 3 4]
         [1 2 3 4]
         [1 2 3 4]]

        >>> a = np.array([1, 2, 3, 4])
        >>> expand_1d_to_2d_array(a, 4, axis=1)
        [[1 1 1 1]
         [2 2 2 2]
         [3 3 3 3]
         [4 4 4 4]]

    :param array:
    :param length:
    :param axis:
    :return:
    """
    from numpy.ma import MaskedArray

    if axis == 0:
        new_shape = (length, array.size)
        reshaped = array
    else:
        new_shape = (array.size, length)
        reshaped = array.reshape(array.size, 1)

    array_2d = np.broadcast_to(reshaped, new_shape, subok=True)

    # Broadcast the mask too (this gets lost otherwise)
    if isinstance(array, MaskedArray):
        array_2d.mask = np.broadcast_to(reshaped.mask, new_shape)

    return array_2d


def create_masked_array_for_missing_data(data, missing_val):
    import numpy.ma as ma
    return ma.array(data, mask=data == missing_val, fill_value=missing_val)


def create_masked_array_for_missing_values(data, missing_values):
    import numpy.ma as ma

    mdata = data
    for missing_value in missing_values:
        if missing_value is not None:
            mdata = ma.masked_where(missing_value == mdata, mdata)

    return mdata


def apply_mask_to_numpy_array(in_array, mask):
    """Element-wise ORs the mask with the mask of the array.
    If the mask masks no elements, no change is made. If the array is not masked, it is converted to a masked array.

    :param in_array: input array
    :type in_array: numpy array or masked array
    :param mask: mask
    :type mask: numpy array of boolean
    """
    import numpy as np

    if mask is not np.ma.nomask and np.any(mask):
        out_array = np.ma.asarray(in_array)
        out_array.mask = out_array.mask | mask
    else:
        out_array = in_array
    return out_array


def array_equal_including_nan(array1, array2):
    """
    :param array1: A numpy array
    :param array2: Another numpy array (can be of a different shape)
    :return: True or false if the arrays are equal, including NaNs.
    """
    import numpy

    try:
        if array1.shape != array2.shape:
            return False
        else:
            for i, j in numpy.nditer([array1, array2]):
                if i != i:
                    pass
                elif i != j:
                    return False
    except AttributeError:
        return False

    return True


def get_coord(data_object, variable, data):
    """
    Find a specified coord

    :param data_object:
    :param variable:
    :param data:
    :return:
    """
    from iris.exceptions import CoordinateNotFoundError
    if variable in [data_object.name(), data_object.standard_name, data_object.long_name, 'default']:
        return data
    else:
        if variable.startswith("search:"):
            number_of_points = float(variable.split(":")[1])
            for coord in data_object.coords():
                if coord.shape[0] == number_of_points:
                    break
        else:
            try:
                coord = data_object.coord(variable)
            except CoordinateNotFoundError:
                return None
        return coord


def fix_longitude_range(lons, range_start):
    """Shifts longitude values by +/- 360 to fit within a 360 degree range starting at a specified value.
    It is assumed that a no shifts larger than 360 are needed.

    :param lons: numpy array of longitude values
    :param range_start: longitude at start of 360 degree range into which values are required to fit
    :return: array of fixed longitudes
    """
    from iris.analysis.cartography import wrap_lons
    return wrap_lons(lons, range_start, 360)


def find_longitude_wrap_start(packed_data_item):
    """
    ONLY WORK OUT THE WRAP START OF THE DATA
    :param packed_data_item:
    :return:
    """
    from iris.exceptions import CoordinateNotFoundError
    try:
        x_points_min = packed_data_item.coord(standard_name='longitude').points.min()
    except CoordinateNotFoundError:
        x_wrap_start = None
    else:
        x_wrap_start = -180 if x_points_min < 0 else 0

    return x_wrap_start


def wrap_longitude_coordinate_values(x_min, x_max):
    if x_min > x_max:
        if x_min >= 180:
            x_min -= 360
        else:
            x_max += 360

    return x_min, x_max


def copy_attributes(source, dest):
    """
    Copy all attributes from one object to another

    :param source: Object to copy attributes from
    :param dest: Object to copy attributes to
    :return: None
    """
    if source:
        if isinstance(source, dict):
            dest.__dict__.update(source)
        else:
            dest.__dict__.update(source.__dict__)


def parse_key_val_string(arguments, separator):
    """
    Takes a (comma) separated list of keyword value pairs (separated by =) and returns a dictionary with those keys and
    values

    :param arguments: A string which is a separated list of keyword value pairs
    :param separator: String which is used to split the string into a list
    :return: A dictionary of the keywords and values
    """
    input_list = arguments.split(separator)
    return parse_key_val_list(input_list)


def parse_key_val_list(input_list):
    """
     Takes list of keyword value strings (seperated by =) and returns a dictionary with those keys and values
     **NOTE** if a key has no value, the key is stored and given the value True

    :param input_list: A list of strings which are keyword value pairs separated by =
    :return: A dictionary of the keywords and values
    """
    key_val_dict = {}
    for element in input_list:
        key_value = element.split('=')
        key = key_value[0]
        value = key_value[1] if len(key_value) > 1 else True
        key_val_dict[key] = value

    return key_val_dict


def haversine(lat, lon, lat2, lon2):
    """
    Computes the Haversine distance between two points
    """
    import math
    R_E = 6378  # Radius of the earth in km
    lat1 = math.radians(lat)
    lat2 = math.radians(lat2)
    lon1 = math.radians(lon)
    lon2 = math.radians(lon2)
    arclen = 2 * math.asin(math.sqrt(
        (math.sin((lat2 - lat1) / 2)) ** 2 + math.cos(lat1) * math.cos(lat2) * (math.sin((lon2 - lon1) / 2)) ** 2))
    return arclen * R_E


class OrderedSet(collections.abc.MutableSet):
    """
    From http://code.activestate.com/recipes/576694/
    """

    def __init__(self, iterable=None):
        self.end = end = []
        end += [None, end, end]  # sentinel node for doubly linked list
        self.map = {}  # key --> [key, prev, next]
        if iterable is not None:
            self |= iterable

    def __len__(self):
        return len(self.map)

    def __contains__(self, key):
        return key in self.map

    def add(self, key):
        if key not in self.map:
            end = self.end
            curr = end[1]
            curr[2] = end[1] = self.map[key] = [key, curr, end]

    def discard(self, key):
        if key in self.map:
            key, prev, next = self.map.pop(key)
            prev[2] = next
            next[1] = prev

    def __iter__(self):
        end = self.end
        curr = end[2]
        while curr is not end:
            yield curr[0]
            curr = curr[2]

    def __reversed__(self):
        end = self.end
        curr = end[1]
        while curr is not end:
            yield curr[0]
            curr = curr[1]

    def pop(self, last=True):
        if not self:
            raise KeyError('set is empty')
        key = self.end[1][0] if last else self.end[2][0]
        self.discard(key)
        return key

    def __repr__(self):
        if not self:
            return '%s()' % (self.__class__.__name__,)
        return '%s(%r)' % (self.__class__.__name__, list(self))

    def __eq__(self, other):
        if isinstance(other, OrderedSet):
            return len(self) == len(other) and list(self) == list(other)
        return set(self) == set(other)


def apply_intersection_mask_to_two_arrays(array1, array2):
    """
    Ensure two (optionally) masked arrays have the same mask.
    If both arrays are masked the intersection of the masks is used.
    If one array is masked and the other is not, the mask from the masked array is applied to the unmasked array.
    If neither array is masked then both arrays are returned as masked arrays with an empty mask.

    :param array1: An (optionally masked) array
    :param array2: Another (optionally masked) array
    :return: Two masked arrays with a common mask
    """
    import numpy.ma as ma
    mask1 = ma.getmaskarray(array1)
    mask2 = ma.getmaskarray(array2)
    # a new masked array combines the masks and preserves the originals
    return ma.masked_array(array1, mask2), ma.masked_array(array2, mask1)


def index_iterator_nditer(shape, points):
    """Iterates over the indexes of a multi-dimensional array of a specified shape.
    The last index changes most rapidly.

    :param shape: sequence of array dimensions
    :return: yields tuples of array indexes
    """

    num_cells = np.product(shape)
    cell_count = 0
    cell_total = 0

    it = np.nditer(points, flags=['multi_index'])
    while not it.finished:
        yield it.multi_index

        # Log progress periodically.
        if cell_count == 10000:
            cell_total += 1
            number_cells_processed = cell_total * 10000
            logging.info("    Processed %d points of %d (%d%%)", number_cells_processed, num_cells,
                         int(number_cells_processed * 100 / num_cells))
            cell_count = 0
        cell_count += 1

        it.iternext()


def index_iterator(shape):
    """Iterates over the indexes of a multi-dimensional array of a specified shape.
    The last index changes most rapidly.

    :param shape: sequence of array dimensions
    :return: yields tuples of array indexes
    """
    dim = len(shape)
    idx = [0] * dim
    num_iterations = 1
    for j in range(0, dim):
        num_iterations *= shape[j]

    for iterations in range(num_iterations):
        yield tuple(idx)
        for j in range(dim - 1, -1, -1):
            idx[j] += 1
            if idx[j] < shape[j]:
                break
            idx[j] = 0


def index_iterator_for_non_masked_data(shape, points):
    """Iterates over the indexes of a multi-dimensional array of a specified shape.
    The last index changes most rapidly.

    :param shape: sequence of array dimensions
    :return: yields tuples of array indexes
    """

    # dim = len(shape)
    # idx = [0] * dim
    # num_iterations = 1
    # for j in range(0, dim):
    #     num_iterations *= shape[j]
    num_cells = np.product(shape)
    cell_count = 0
    cell_total = 0

    it = np.nditer(points.data, flags=['multi_index'])
    while not it.finished:
        if it[0] is not np.ma.masked:
            yield it.multi_index

        it.iternext()

        # Log progress periodically.
        if cell_count == 10000:
            cell_total += 1
            number_cells_processed = cell_total * 10000
            logging.info("    Processed %d points of %d (%d%%)", number_cells_processed, num_cells,
                         int(number_cells_processed * 100 / num_cells))
            cell_count = 0
        cell_count += 1


def parse_distance_with_units_to_float_km(distance):
    """Parse a string such as '10km' or '1.0e3m' to a distance in km

    :param distance: string to parse
    :return: A distance in km
    """
    measurement = split_into_float_and_units(distance)
    distance_in_km = measurement['value']
    if measurement['units'] == 'm':
        distance_in_km /= 1000.0
    elif measurement['units'] == 'km' or measurement['units'] is None:
        pass
    else:
        raise InvalidCommandLineOptionError("Units not recognised", measurement['units'])

    return distance_in_km


def parse_distance_with_units_to_float_m(distance):
    """Parse a string such as '10km' or '1.0e3m' to a distance in m

    :param distance: string to parse
    :return: A distance in m
    """
    measurement = split_into_float_and_units(distance)
    if measurement['units'] is None:
        return measurement['value']
    else:
        return parse_distance_with_units_to_float_km(distance) * 1000.0


def split_into_float_and_units(measurement):
    """Split a string such as '1000m' or '1.0e3' to a value and, optionally, units

    :param distance: string to parse
    :return: A distance in m
    """

    # Convert to a string, just in case the number comes in as something else
    measurement = str(measurement)

    # First find any numbers that match the float type, e.g. 1, 1.2, 12e3
    measurement_value = re.findall(r'[-+]?(?:\d+(?:\.\d*)?|\.\d+)(?:[eE][-+]?\d+)?', measurement)

    # Check for 1 and only 1 value found
    if len(measurement_value) > 1:
        raise InvalidCommandLineOptionError("More than one number found")
    elif len(measurement_value) == 0:
        raise InvalidCommandLineOptionError("No numbers found")

    # Remove number from string, so floats such as 1e3 do not have the e confused for a unit
    measurement = measurement.replace(measurement_value[0], " ")

    # Match any strings of characters, which should be units
    measurement_units = re.findall(r'[a-zA-z]+', measurement)

    if len(measurement_units) > 1:
        raise InvalidCommandLineOptionError("More than one unit found")
    elif len(measurement_units) == 0:
        measurement_units = [None]

    return {'value': float(measurement_value[0]), 'units': measurement_units[0]}


def get_class_name(cls):
    """Returns the qualified class name of a class.

    :param cls: class
    :return: class name
    """
    return cls.__module__ + '.' + cls.__name__


def isnan(number):
    return number != number

standard_names = {'longitude': 'X', 'grid_longitude': 'X', 'projection_x_coordinate': 'X',
                  'latitude': 'Y', 'grid_latitude': 'Y', 'projection_y_coordinate': 'Y',
                  'altitude': 'Z', 'time': 'T', 'air_pressure': 'P'}

# The standard axes lookup is just the inverse of the standard names with the less likely options removed to preserve
# a one to one mapping
standard_axes = {'X': 'longitude', 'Y': 'latitude', 'Z': 'altitude', 'T': 'time', 'P': 'air_pressure'}


def guess_coord_axis(coord):
    """Returns X, Y, Z or T corresponding to longitude, latitude,
    altitude or time respectively if the coordinate can be determined
    to be one of these (based on the standard name only, in this implementation).

    This is intended to be similar to iris.util.guess_coord_axis.
    """
    import iris
    # TODO Can more be done for ungridded based on units, as with iris.util.guess_coord_axis?
    guessed_axis = None
    if isinstance(coord, iris.coords.Coord):
        guessed_axis = iris.util.guess_coord_axis(coord)
    elif coord.standard_name is not None:
        guessed_axis = standard_names.get(coord.standard_name.lower())
    return guessed_axis


def add_to_list_if_not_none(item, list):
    """
    Add a value to a list if it is not None

    :param item: the item to add
    :param list: the list to append it to
    :return: nothing
    """
    if item is not None:
        list.append(item)


def dimensions_compatible(dimensions, other_dimensions):
    """
    Check to see if two dimensions share the same common coordinates. Note that this will only compare the dimensions
    up to the length of the shortest list.

    :param dimensions: dimension list
    :param other_dimensions: other dimension list
    """

    for dim, other_dim in zip(dimensions, other_dimensions):
        if not dim == other_dim:
            return False
    return True


def set_standard_name_if_valid(data, standard_name):
    """
    Set a data object's standard name if it is a valid CF compliant name, otherwise set it to None

    :param CommonData or Metadata data: Data to set standard name on
    :param standard_name: Standard name to set
    :return:
    """
    try:
        data.standard_name = standard_name
    except ValueError:
        logging.info("Not setting standard name '{}' as it is not CF compliant.".format(standard_name))
        data.standard_name = None


def deprecated(func):
    """
    This is a decorator which can be used to mark functions
    as deprecated. It will result in a warning being emitted
    when the function is used.

    Taken from http://code.activestate.com/recipes/391367-deprecated/
    """

    def newFunc(*args, **kwargs):
        def log_warning(message, category, filename, lineno, file=None):
            logging.warning('%s:%s %s:%s' % (category.__name__, message, filename, lineno))

        warnings.showwarning = log_warning
        warnings.simplefilter('once', DeprecationWarning)  # turn off filter
        warnings.warn("Call to deprecated function %s." % func.__name__,
                      category=DeprecationWarning, stacklevel=2)
        return func(*args, **kwargs)

    newFunc.__name__ = func.__name__
    newFunc.__doc__ = func.__doc__
    newFunc.__dict__.update(func.__dict__)
    return newFunc


def listify(item):
    """
    If item is not a list, return it as a list

    :param item: Item which may or may not be a list
    :return: List
    """
    from types import GeneratorType
    if isinstance(item, (tuple, GeneratorType)):
        return list(item)
    if not isinstance(item, list):
        return [item]
    return item


def log_memory_profile(location):
    """
    Write the total memory to the log as debug message

    :param location: location in the program where the memory measurement was taken
    :return: nothing
    """
    try:
        import psutil
        mem = psutil.Process().memory_info().rss / BYTES_IN_A_MB
        logging.debug("App Memory MB ({}): {}".format(location, mem))
    except ImportError:
        pass


def move_item_to_end(iter, item):
    """
    Move an item in an iterable to the end of a list
    :param iterable iter: iterable container (list or tuple) contianing item
    :param item: item to move to end
    :return list: rearranged list
    """
    if not isinstance(iter, list):
        iter = list(iter)
    dim = iter.pop(iter.index(item))
    iter.append(dim)
    return iter


@contextlib.contextmanager
def demote_warnings(level=logging.INFO):
    import warnings
    with warnings.catch_warnings(record=True) as ws:
        warnings.simplefilter("always")
        yield
        for w in ws:
            logging.log(level, w.message)


@contextlib.contextmanager
def single_warnings_only():
    import warnings
    with warnings.catch_warnings():
        warnings.simplefilter("once")
        yield


def squeeze(data):
    from iris.cube import Cube
    from iris.util import squeeze
    from cis.data_io.gridded_data import make_from_cube
    if isinstance(data, Cube):
        return make_from_cube(squeeze(data))
    else:
        return data


@contextlib.contextmanager
def no_autoscale(ax):
    # Turn the scaling off so that we don't change the limits by doing whatever we do next
    ax.set_autoscale_on(False)
    yield
    # Turn scaling back on
    ax.set_autoscale_on(True)



import unittest

from hamcrest import assert_that, is_, close_to

from cis import read_data
from cis.test.integration_test_data import *


class TestNetCDF_CF_Gridded_ECHAM_Hybrid_Pressure_Geopotential(unittest.TestCase):

    data = None

    def read_data(self):
        # Only do this once
        if self.data is None:
            filename = valid_echamham_geopotential_filename
            var = valid_echamham_geopotential_variable
            self.data = read_data(filename, var)

    def test_GIVEN_hybrid_pressure_readable_by_iris_WHEN_read_THEN_pressure_coordinate_present(self):
        self.read_data()
        # The air pressure aux coord is a CF compliant aux coord which IRIS identifies
        pressure = self.data.coord('air_pressure')
        assert_that(pressure.shape, is_((248, 31, 96, 192)))

    def test_GIVEN_altitude_not_read_by_iris_WHEN_read_THEN_pressure_coordinate_present(self):
        self.read_data()
        # The altitude aux coord is CF compliant but not identified by IRIS so we manually create it
        altitude = self.data.coord('altitude')
        assert_that(altitude.shape, is_((248, 31, 96, 192)))
        # Check it has been converted to meters
        assert_that(str(altitude.units), is_('meter'))
        assert_that(altitude.points[0, 0, 0, 0], close_to(275081.0 / 9.80665, 0.1))


class TestNetCDF_CF_Gridded_ECHAM_Hybrid_Pressure_Geopotential_Height(unittest.TestCase):

    data = None

    def read_data(self):
        # Only need to do this once
        if self.data is None:
            filename = valid_echamham_geopotential_height_filename
            var = valid_echamham_geopotential_height_variable
            self.data = read_data(filename, var)

    def test_GIVEN_hybrid_pressure_readable_by_iris_WHEN_read_THEN_pressure_coordinate_present(self):
        self.read_data()
        # The air pressure aux coord is a CF compliant aux coord which IRIS identifies
        pressure = self.data.coord('air_pressure')
        assert_that(pressure.shape, is_((248, 31, 96, 192)))

    def test_GIVEN_altitude_not_read_by_iris_WHEN_read_THEN_pressure_coordinate_present(self):
        self.read_data()
        # The altitude aux coord is CF compliant but not identified by IRIS so we manually create it
        altitude = self.data.coord('altitude')
        assert_that(altitude.shape, is_((248, 31, 96, 192)))

if __name__ == '__main__':
    unittest.main()



from abc import ABCMeta, abstractmethod
import six
import logging

import numpy as np
import iris
import iris.coords

import cis.data_io.gridded_data as gridded_data


def subset(data, constraint, **kwargs):
    """
    Helper function for constraining a CommonData or CommonDataList object (data) given a SubsetConstraint
    class, the constraints should be specified using the kwargs of the form coord: [min, max]

    A shape keyword can also be supplied as a WKT string or shapely object to subset in lat/lon by an arbitrary shape.

    :param CommonData or CommonDataList data: The data to subset
    :param class SubsetConstraint constraint: A SubsetConstraint class to do the constraining
    :param kwargs: The limits as slices or length 2 tuples of max and min.
    :return:
    """
    from cis.time_util import PartialDateTime
    from cis.exceptions import CoordinateNotFoundError
    from shapely.wkt import loads

    constraints = {}

    for dim_name, limit in kwargs.items():
        # Deal with shape argument
        if dim_name == 'shape':
            if isinstance(limit, six.string_types):
                try:
                    shape = loads(limit)
                except Exception:
                    raise ValueError("Invalid shape string: " + limit)
            else:
                shape = limit
            constraints['shape'] = shape
            bounding_box = shape.bounds
            # Create the lat/lon box - this will be used to speed up the shape subset
            constraints[data.coord(standard_name='longitude').name()] = slice(bounding_box[0], bounding_box[2])
            constraints[data.coord(standard_name='latitude').name()] = slice(bounding_box[1], bounding_box[3])
            continue

        c = data._get_coord(dim_name)
        if c is None:
            raise CoordinateNotFoundError("No coordinate found that matches '{}'. Please check the "
                                          "coordinate name.".format(dim_name))

        if all(hasattr(limit, att) for att in ('start', 'stop')):
            l = limit
        elif isinstance(limit, PartialDateTime):
            l = slice(limit.min(), limit.max())
        elif len(limit) == 1 and isinstance(limit[0], PartialDateTime):
            l = slice(limit[0].min(), limit[0].max())
        elif len(limit) == 2:
            l = slice(limit[0], limit[1])
        else:
            raise ValueError("Invalid subset arguments: {}".format(limit))

        # Fill in defaults
        limit_start = l.start if l.start is not None else c.points.min()
        limit_end = l.stop if l.stop is not None else c.points.max()

        constraints[data._get_coord(dim_name).name()] = slice(limit_start, limit_end)

    subset_constraint = constraint(constraints)

    subset = subset_constraint.constrain(data)

    if subset is not None:
        subset.add_history("Subsetted using limits: " + str(subset_constraint))

    return subset


@six.add_metaclass(ABCMeta)
class SubsetConstraint(object):
    """Abstract Constraint for subsetting.

    Holds the limits for subsetting in each dimension.
    """

    def __init__(self, limits, shape=None):
        """
        :param dict limits: A dictionary mapping coordinate name to slice objects
        """
        self._limits = limits
        self._shape = shape
        logging.debug("Created SubsetConstraint of type %s", self.__class__.__name__)

    def __str__(self):
        limit_strs = []
        for name, limit in self._limits.items():
            limit_strs.append("{}: [{}, {}]".format(name, str(limit.start), str(limit.stop)))
        return ', '.join(limit_strs)

    @abstractmethod
    def constrain(self, data):
        """Subsets the supplied data.

        :param data: data to be subsetted
        :return: subsetted data
        """


class GriddedSubsetConstraint(SubsetConstraint):
    """
    Implementation of SubsetConstraint for subsetting gridded data.
    """

    def constrain(self, data):
        """
        Subsets the supplied data using a combination of iris.cube.Cube.extract and iris.cube.Cube.intersection,
        depending on whether intersection is supported (whether the coordinate has a defined modulus).
        :param data: data to be subsetted
        :return: subsetted data or None if all data excluded.
        @rtype: cis.data_io.gridded_data.GriddedData
        """
        _shape = self._limits.pop('shape', None)
        extract_constraint, intersection_constraint = self._make_extract_and_intersection_constraints(data)
        if extract_constraint is not None:
            data = data.extract(extract_constraint)
            if data is None:
                return None  # Don't do the intersection
        if intersection_constraint:
            try:
                data = data.intersection(**intersection_constraint)
            except IndexError:
                return None

        if _shape is not None:
            if data.ndim > 2:
                raise NotImplementedError("Unable to perform shape subset for multidimensional gridded datasets")
            mask = np.ones(data.shape, dtype=bool)
            mask[np.unravel_index(_get_gridded_subset_region_indices(data, _shape), data.shape)] = False
            if isinstance(data.data, np.ma.MaskedArray):
                data.data.mask &= mask
            else:
                data.data = np.ma.masked_array(data.data, mask)
        return gridded_data.make_from_cube(data)

    def _make_extract_and_intersection_constraints(self, data):
        """
        Make the appropriate constraints:
        - dictionary of coord_name -> (min, max) for coordinates with defined modulus (to be used on the IRIS
          intersection method).
        - iris.Constraint if no defined modulus
        :param data:
        :return:
        """
        class Contains(object):
            """
            Callable object for checking if a value is within a preset range
            """
            def __init__(self, lower, upper):
                self.lower = lower
                self.upper = upper

            def __call__(self, x):
                return self.lower <= x.point <= self.upper

        intersection_constraint = {}
        extract_constraint = iris.Constraint()
        for coord, limits in self._limits.items():
            if data.coord(coord).units.modulus is not None:
                # These coordinates can be safely used with iris.cube.Cube.intersection()
                intersection_constraint[coord] = (limits.start, limits.stop)
            else:
                # These coordinates cannot be used with iris.cube.Cube.intersection(), will use iris.cube.Cube.extract()
                extract_constraint = extract_constraint & iris.Constraint(**{coord: Contains(limits.start, limits.stop)})
        return extract_constraint, intersection_constraint


class UngriddedSubsetConstraint(SubsetConstraint):
    """
    Implementation of SubsetConstraint for subsetting ungridded data.
    """
    def __init__(self, limits):
        super(UngriddedSubsetConstraint, self).__init__(limits)
        self._shape_indices = None
        self._combined_mask = None

    def constrain(self, data):
        """Subsets the supplied data.

        :param data: data to be subsetted
        :return: subsetted data
        """
        import numpy as np
        from datetime import datetime
        from cis.data_io.ungridded_data import UngriddedDataList

        if isinstance(data, list):
            # Calculating masks and indices will only take place on the first iteration,
            # so we can just call this method recursively if we've got a list of data.
            output = UngriddedDataList()
            for var in data:
                output.append(self.constrain(var))
            return output

        _data = self._create_data_for_subset(data)

        _shape = self._limits.pop('shape', None)

        if self._combined_mask is None:
            # Create the combined mask across all limits
            shape = _data.coords()[0].data.shape  # This assumes they are all the same shape
            combined_mask = np.ones(shape, dtype=bool)
            for coord, limit in self._limits.items():
                # Convert the points to datetimes if the limit is a datetime
                if isinstance(limit.start, datetime):
                    points = _data.coord(coord).units.num2date(_data.coord(coord).data)
                else:
                    points = _data.coord(coord).data
                # Select any points which are <= to the stop limit AND >= to the start limit
                mask = (np.less_equal(points, limit.stop) & np.greater_equal(points, limit.start))
                combined_mask &= mask
            self._combined_mask = combined_mask

        _data = _data[self._combined_mask]

        if _shape is not None:
            if self._shape_indices is None:
                self._shape_indices = _get_ungridded_subset_region_indices(_data, _shape)
            _data = _data[np.unravel_index(self._shape_indices, _data.shape)]

        if _data.size == 0:
            _data = None

        return _data

    def _create_data_for_subset(self, data):
        """
        Produce a copy of the data so that the original is not altered,
        with longitudes mapped onto the right domain for the requested limits
        :param data: Data being subsetted
        :return:
        """
        from cis.exceptions import CoordinateNotFoundError
        # We need a copy with new data and coordinates
        data = data.copy()
        # Check for longitude coordinate in the limits
        for dim_name, limit in self._limits.items():
            try:
                coord = data.coord(dim_name)
            except CoordinateNotFoundError:
                # E.g. shape...
                continue
            if coord.standard_name == 'longitude':
                coord_min = coord.points.min()
                coord_max = coord.points.max()
                data_below_zero = coord_min < 0
                data_above_180 = coord_max > 180
                limits_below_zero = limit.start < 0 or limit.stop < 0
                limits_above_180 = limit.start > 180 or limit.stop > 180

                if data_below_zero and not data_above_180:
                    # i.e. data is in the range -180 -> 180
                    # Only convert the data if the limits are above 180:
                    if limits_above_180 and not limits_below_zero:
                        # Convert data from -180 -> 180 to 0 -> 360
                        range_start = 0
                        coord.set_longitude_range(range_start)
                elif data_above_180 and not data_below_zero:
                    # i.e. data is in the range 0 -> 360
                    if limits_below_zero and not limits_above_180:
                        # Convert data from 0 -> 360 to -180 -> 180
                        range_start = -180
                        coord.set_longitude_range(range_start)
        return data


def _get_indices_for_lat_lon_points(lats, lons, region):
    from shapely.geometry import MultiPoint

    lat_lon_points = np.vstack([lats, lons])
    points = MultiPoint(lat_lon_points.T)

    # Performance in this loop might be an issue, but I think it's essentially how GeoPandas does it. If I want to
    #  improve it I might need to look at using something like rtree.
    return [i for i, p in enumerate(points) if region.contains(p)]


def _get_ungridded_subset_region_indices(ungridded_data, region):
    # We have to use flatten rather than flat, GEOS creates a copy of the data if it's a view anyway.
    return _get_indices_for_lat_lon_points(ungridded_data.lon.data.flatten(), ungridded_data.lat.data.flatten(), region)


def _get_gridded_subset_region_indices(gridded_data, region):
    # Using X and Y is a bit more general than lat and lon - the shapefiles needn't actually represent lat/lon
    x, y = np.meshgrid(gridded_data.coord(axis='X').points, gridded_data.coord(axis='Y').points)
    return _get_indices_for_lat_lon_points(x.flat, y.flat, region)
